import datetime
import glob
import json
import logging
import os
import random
import shutil
import warnings
from concurrent.futures import ThreadPoolExecutor, as_completed
from pathlib import Path

import geopandas as gpd
import osmnx as ox
import pandas as pd
import requests
from PIL import Image
from shapely.errors import ShapelyDeprecationWarning
from tqdm import tqdm

import zensvi.download.mapillary.interface as mly
from zensvi.download.base import BaseDownloader
from zensvi.download.utils.helpers import check_and_buffer, standardize_column_names
from zensvi.utils.log import Logger

warnings.filterwarnings("ignore", category=ShapelyDeprecationWarning)
logging.getLogger("mapillary.utils.client").setLevel(logging.WARNING)


class MLYDownloader(BaseDownloader):
    """Mapillary Downloader class.

    Args:
        mly_api_key (str, optional): Mapillary API key. Defaults to None.
        log_path (str, optional): Path to the log file. Defaults to None.
        max_workers (int, optional): Number of workers for parallel processing. Defaults to None.
    """

    def __init__(self, mly_api_key, log_path=None, max_workers=None):
        super().__init__(log_path)
        self._mly_api_key = mly_api_key
        self._max_workers = max_workers
        mly.set_access_token(self.mly_api_key)
        # initialize the logger
        if log_path is not None:
            self.logger = Logger(log_path)
        else:
            self.logger = None

    @property
    def mly_api_key(self):
        """Property for Mapillary API key.

        Returns:
            str: mly_api_key
        """
        return self._mly_api_key

    @mly_api_key.setter
    def mly_api_key(self, mly_api_key):
        self._mly_api_key = mly_api_key

    @property
    def max_workers(self):
        """Property for the number of workers for parallel processing.

        Returns:
            int: max_workers
        """
        return self._max_workers

    @max_workers.setter
    def max_workers(self, max_workers):
        if max_workers is None:
            self._max_workers = min(32, os.cpu_count() + 4)
        else:
            self._max_workers = max_workers

    def _read_pids(self, path_pid):
        pid_df = pd.read_csv(path_pid)
        # drop NA values in id columns
        pid_df = pid_df.dropna(subset=["id"])
        # get unique pids (ie "id" columns) as a list
        pids = pid_df["id"].astype("int64").unique().tolist()
        return pids

    def _set_dirs(self, dir_output):
        # set dir_output as attribute and create the directory
        self.dir_output = Path(dir_output)
        self.dir_output.mkdir(parents=True, exist_ok=True)
        self.pids_url = self.dir_output / "pids_urls.csv"
        # set dir_cache as attribute and create the directory
        self.dir_cache = self.dir_output / "cache_zensvi"
        self.dir_cache.mkdir(parents=True, exist_ok=True)
        # set other cache directories
        self.cache_lat_lon = self.dir_cache / "lat_lon.csv"
        self.cache_pids_raw = self.dir_cache / "pids_raw.csv"

    def _get_pids_from_gdf(self, gdf, mly_kwargs, **kwargs):
        # set crs to EPSG:4326 if it's None
        if gdf.crs is None:
            gdf = gdf.set_crs("EPSG:4326")
        elif gdf.crs != "EPSG:4326":
            # convert to EPSG:4326
            gdf = gdf.to_crs("EPSG:4326")

        # convert to FeatureCollection
        geojson = json.loads(gdf.to_json())

        # use get pids with mly.interface.images_in_geojson
        if kwargs["use_cache"]:
            result_json = mly.images_in_geojson(
                geojson,
                dir_cache=self.dir_cache,
                max_workers=self.max_workers,
                logger=self.logger,
                **mly_kwargs,
            ).to_dict()["features"]
        else:
            result_json = mly.images_in_geojson(
                geojson, max_workers=self.max_workers, logger=self.logger, **mly_kwargs
            ).to_dict()["features"]

        # convert to geodataframe
        result_gdf = gpd.GeoDataFrame.from_features(result_json)

        # return None if there is no result
        if len(result_gdf) == 0:
            return None

        # add lon and lat columns
        result_gdf["lon"] = result_gdf["geometry"].apply(lambda geom: geom.x)
        result_gdf["lat"] = result_gdf["geometry"].apply(lambda geom: geom.y)

        # drop geometry column
        result_df = result_gdf.drop(columns="geometry")

        return result_df

    def _get_raw_pids(self, **kwargs):
        mly_allowed_keys = {
            "compass_angle",
            "image_type",
            "min_captured_at",
            "max_captured_at",
            "organization_id",
        }
        mly_kwargs = {k: v for k, v in kwargs.items() if k in mly_allowed_keys}
        if self.cache_pids_raw.exists():
            pid = pd.read_csv(self.cache_pids_raw)
            print("The raw panorama IDs have been read from the cache")
            return pid

        # input: lat and lon
        if kwargs["lat"] is not None and kwargs["lon"] is not None:
            # create a geodataframe with lat and lon
            df = pd.DataFrame({"lat": [kwargs["lat"]], "lon": [kwargs["lon"]]})
            gdf = gpd.GeoDataFrame(df, geometry=gpd.points_from_xy(df.lon, df.lat), crs="EPSG:4326")

        # input: csv file
        elif kwargs["input_csv_file"] != "":
            df = pd.read_csv(kwargs["input_csv_file"])
            df = standardize_column_names(df)
            gdf = gpd.GeoDataFrame(
                df,
                geometry=gpd.points_from_xy(df.longitude, df.latitude),
                crs="EPSG:4326",
            )

        # input: shapefile
        elif kwargs["input_shp_file"] != "":
            gdf = gpd.read_file(kwargs["input_shp_file"])

        # input: place name
        elif kwargs["input_place_name"] != "":
            print("Geocoding the input place name")
            gdf = ox.geocoder.geocode_to_gdf(kwargs["input_place_name"])
            # raise error if the input_place_name is not found
            if len(gdf) == 0:
                raise ValueError("The input_place_name is not found. Please try another place name.")
        else:
            raise ValueError("Please input the lat and lon, csv file, or shapefile.")

        # check geometry type and buffer
        gdf = check_and_buffer(gdf, kwargs["buffer"])
        # get pid
        pid = self._get_pids_from_gdf(gdf, mly_kwargs, **kwargs)

        return pid

    def _filter_pids_date(self, pid_df, start_date, end_date):
        # create a temporary column date from captured_at (milliseconds from Unix epoch)
        pid_df["date"] = pd.to_datetime(pid_df["captured_at"], unit="ms")
        # check if start_date and end_date are in the correct format with regex. If not, raise error
        if start_date is not None:
            try:
                start_date = datetime.datetime.strptime(start_date, "%Y-%m-%d")
            except ValueError:
                raise ValueError("Incorrect start_date format, should be YYYY-MM-DD")
        if end_date is not None:
            try:
                end_date = datetime.datetime.strptime(end_date, "%Y-%m-%d")
            except ValueError:
                raise ValueError("Incorrect end_date format, should be YYYY-MM-DD")
        # if start_date is not None, filter out the rows with date < start_date
        pid_df = pid_df[pid_df["date"] >= start_date] if start_date is not None else pid_df
        # if end_date is not None, filter out the rows with date > end_date
        pid_df = pid_df[pid_df["date"] <= end_date] if end_date is not None else pid_df
        # drop the temporary column date
        pid_df = pid_df.drop(columns="date")
        return pid_df

    def _get_pids(self, path_pid, **kwargs):
        # get raw pid
        pid = self._get_raw_pids(**kwargs)

        if pid is None or len(pid) == 0:
            print("There is no panorama ID to download")
            return

        # # Assuming that df is your DataFrame and that 'geometry' column contains Point objects
        # # convert the geometry column into shapely geometry objects
        # pid['geometry'] = pid['geometry'].apply(wkt.loads)
        # pid['lon'] = pid['geometry'].apply(lambda geom: geom.x)
        # pid['lat'] = pid['geometry'].apply(lambda geom: geom.y)
        # # drop geometry column
        # pid = pid.drop(columns='geometry')
        # move the "id" column to the first column
        pid = pid[["id"] + [col for col in pid.columns if col != "id"]]

        # keep id,captured_at,compass_angle,is_pano,sequence_id,lon,lat, and organization_id if it exists
        columns_to_keep = [
            "id",
            "captured_at",
            "compass_angle",
            "creator_id",
            "is_pano",
            "sequence_id",
            "lon",
            "lat",
        ]

        # Add organization_id to columns_to_keep if it exists in the DataFrame
        if "organization_id" in pid.columns:
            columns_to_keep.append("organization_id")

        pid = pid[columns_to_keep]

        pid.to_csv(path_pid, index=False)
        print("The panorama IDs have been saved to {}".format(path_pid))

        # if use_cache, delete self.dir_cache / "tiles_results" directory
        if kwargs["use_cache"]:
            dir_cache_tiles = self.dir_cache / "tiles_results"
            if dir_cache_tiles.exists():
                shutil.rmtree(dir_cache_tiles)
                print("The cache directory for tiles has been deleted")

    def _get_urls_mly(self, path_pid, resolution=1024, additional_fields=["all"]):
        # check if seld.cache_pids_urls exists
        if self.pids_url.exists():
            print("The panorama URLs have been read from the cache")
            return

        dir_cache_urls = self.dir_cache / "urls"
        dir_cache_urls.mkdir(parents=True, exist_ok=True)

        checkpoints = glob.glob(str(dir_cache_urls / "*.csv"))
        checkpoint_start_index = len(checkpoints)

        panoids = set(self._read_pids(path_pid))  # Convert to set for faster operations
        if len(panoids) == 0:
            print("There is no panorama ID to download")
            return

        # Read all panoids from the checkpoint files
        completed_panoids = set()  # Use set for faster operations
        for checkpoint in checkpoints:
            try:
                df_checkpoint = pd.read_csv(checkpoint)
                completed_panoids.update(df_checkpoint["id"].tolist())
            except pd.errors.EmptyDataError:
                print(f"Warning: {checkpoint} is empty and has been skipped.")
                continue

        # Filter out the panoids that have already been processed
        panoids = list(panoids - completed_panoids)  # Subtract sets and convert back to list
        if len(panoids) == 0:
            print("All images have been downloaded")
            return

        def worker(panoid, resolution):
            result = mly.image_thumbnail(panoid, resolution=resolution, additional_fields=additional_fields)
            return panoid, result

        results = {}
        batch_size = 1000  # Modify this to a suitable value
        num_batches = (len(panoids) + batch_size - 1) // batch_size

        for i in tqdm(
            range(num_batches),
            desc=f"Getting urls by batch size {min(batch_size, len(panoids))}",
        ):
            with ThreadPoolExecutor() as executor:
                batch_futures = {
                    executor.submit(worker, panoid, resolution): panoid
                    for panoid in panoids[i * batch_size : (i + 1) * batch_size]
                }

                for future in tqdm(
                    as_completed(batch_futures),
                    total=len(batch_futures),
                    desc=f"Getting urls for batch #{i+1}",
                ):
                    current_panoid = batch_futures[future]
                    try:
                        panoid, result = future.result()
                        results[panoid] = result
                    except Exception as e:
                        print(f"Error: {e}")
                        if self.logger is not None:
                            self.logger.log_failed_pids(current_panoid)
                        continue

            if len(results) > 0:
                df = (
                    pd.DataFrame.from_dict(results, orient="index")
                    .reset_index(drop=True)
                    .rename(columns={f"thumb_{resolution}_url": "url"})
                )
                df = df[["id"] + [col for col in df.columns if col != "id"]]
                df.to_csv(
                    f"{dir_cache_urls}/checkpoint_batch_{checkpoint_start_index+i+1}.csv",
                    index=False,
                )
            results = {}

        # Merge all checkpoints into a single dataframe
        results_df = pd.concat(
            [pd.read_csv(checkpoint) for checkpoint in glob.glob(str(dir_cache_urls / "*.csv"))],
            ignore_index=True,
        )
        results_df.to_csv(self.pids_url, index=False)

        if dir_cache_urls.exists():
            shutil.rmtree(dir_cache_urls)

    def _download_images_mly(self, path_pid, cropped, batch_size, start_date, end_date):
        checkpoints = glob.glob(str(self.panorama_output / "**/*.png"), recursive=True)

        # Read already downloaded images and convert to ids
        downloaded_ids = set([Path(file_path).stem for file_path in checkpoints])  # Use set for faster operations

        pid_df = pd.read_csv(path_pid).dropna(subset=["id"])
        pid_df["id"] = pid_df["id"].astype("int64")
        urls_df = pd.read_csv(self.pids_url)
        urls_df["id"] = urls_df["id"].astype("int64")
        # merge pid_df and urls_df, keeping only one instance of duplicated columns
        urls_df = urls_df.merge(pid_df, on="id", how="left", suffixes=("", "_drop"))
        urls_df = urls_df.loc[:, ~urls_df.columns.str.endswith("_drop")]
        # filter out the rows by date
        urls_df = self._filter_pids_date(urls_df, start_date, end_date)

        # Filter out the ids that have already been processed
        urls_df = urls_df[~urls_df["id"].isin(downloaded_ids)]  # Use isin for efficient operation

        def worker(row, output_dir, cropped):
            url, panoid = row.url, row.id
            user_agent = random.choice(self._user_agents)
            proxy = random.choice(self._proxies)

            image_name = f"{panoid}.png"  # Use id for file name
            image_path = output_dir / image_name
            try:
                response = requests.get(url, headers=user_agent, proxies=proxy, timeout=10)
                if response.status_code == 200:
                    with open(image_path, "wb") as f:
                        f.write(response.content)

                    if cropped:
                        img = Image.open(image_path)
                        w, h = img.size
                        img_cropped = img.crop((0, 0, w, h // 2))
                        img_cropped.save(image_path)

                else:
                    if self.logger is not None:
                        self.logger.log_failed_pids(panoid)
            except Exception as e:
                if self.logger is not None:
                    self.logger.log_failed_pids(panoid)
                print(f"Error: {e}")

        num_batches = (len(urls_df) + batch_size - 1) // batch_size

        # Calculate current highest batch number
        existing_batches = glob.glob(str(self.panorama_output / "batch_*"))
        existing_batch_numbers = [int(Path(batch).name.split("_")[-1]) for batch in existing_batches]
        start_batch_number = max(existing_batch_numbers, default=0)

        for i in tqdm(
            range(start_batch_number, start_batch_number + num_batches),
            desc=f"Downloading images by batch size {min(batch_size, len(urls_df))}",
        ):
            # Create a new sub-folder for each batch
            batch_out_path = self.panorama_output / f"batch_{i+1}"
            batch_out_path.mkdir(exist_ok=True)

            with ThreadPoolExecutor() as executor:
                batch_futures = {
                    executor.submit(worker, row, batch_out_path, cropped): row.id
                    for row in urls_df.iloc[i * batch_size : (i + 1) * batch_size].itertuples()
                }
                for future in tqdm(
                    as_completed(batch_futures),
                    total=len(batch_futures),
                    desc=f"Downloading images for batch #{i+1}",
                ):
                    try:
                        future.result()
                    except Exception as e:
                        print(f"Error: {e}")

    def download_svi(
        self,
        dir_output,
        path_pid=None,
        lat=None,
        lon=None,
        input_csv_file="",
        input_shp_file="",
        input_place_name="",
        buffer=0,
        update_pids=False,
        resolution=1024,
        cropped=False,
        batch_size=1000,
        start_date=None,
        end_date=None,
        metadata_only=False,
        use_cache=True,
        additional_fields=["all"],
        **kwargs,
    ):
        """Downloads street view images from Mapillary using specified parameters.

        Args:
            dir_output (str): Directory where output files and images will be stored.
            path_pid (str, optional): Path to a file containing panorama IDs. If not provided, IDs will be fetched based on other parameters.
            lat (float, optional): Latitude to fetch panorama IDs around this point. Must be used with `lon`.
            lon (float, optional): Longitude to fetch panorama IDs around this point. Must be used with `lat`.
            input_csv_file (str, optional): Path to a CSV file containing locations for which to fetch panorama IDs.
            input_shp_file (str, optional): Path to a shapefile containing geographic locations for fetching panorama IDs.
            input_place_name (str, optional): A place name for geocoding to fetch panorama IDs.
            buffer (int, optional): Buffer size in meters to expand the geographic area for panorama ID fetching.
            update_pids (bool, optional): If True, will update panorama IDs even if a valid `path_pid` is provided. Defaults to False.
            resolution (int, optional): The resolution of the images to download. Defaults to 1024.
            cropped (bool, optional): If True, images will be cropped to the upper half. Defaults to False.
            batch_size (int, optional): Number of images to process in each batch. Defaults to 1000.
            start_date (str, optional): Start date (YYYY-MM-DD) to filter images by capture date.
            end_date (str, optional): End date (YYYY-MM-DD) to filter images by capture date.
            metadata_only (bool, optional): If True, skips downloading images and only fetches metadata. Defaults to False.
            use_cache (bool, optional): If True, uses cached data to speed up the operation. Defaults to True.
            additional_fields (list, optional): Additional fields to fetch from the API. Defaults to ["all"].
                Possible fields include:
                1. altitude - float, original altitude from Exif
                2. atomic_scale - float, scale of the SfM reconstruction around the image
                3. camera_parameters - array of float, intrinsic camera parameters
                4. camera_type - enum, type of camera projection (perspective, fisheye, or spherical)
                5. captured_at - timestamp, capture time
                6. compass_angle - float, original compass angle of the image
                7. computed_altitude - float, altitude after running image processing
                8. computed_compass_angle - float, compass angle after running image processing
                9. computed_geometry - GeoJSON Point, location after running image processing
                10. computed_rotation - enum, corrected orientation of the image
                11. exif_orientation - enum, orientation of the camera as given by the exif tag
                12. geometry - GeoJSON Point geometry
                13. height - int, height of the original image uploaded
                14. thumb_256_url - string, URL to the 256px wide thumbnail
                15. thumb_1024_url - string, URL to the 1024px wide thumbnail
                16. thumb_2048_url - string, URL to the 2048px wide thumbnail
                17. merge_cc - int, id of the connected component of images that were aligned together
                18. mesh - { id: string, url: string } - URL to the mesh
                19. quality_score - float, how good the image is (experimental)
                20. sequence - string, ID of the sequence
                21. sfm_cluster - { id: string, url: string } - URL to the point cloud
                22. width - int, width of the original image uploaded
            **kwargs: Additional keyword arguments that are passed to the API.

        Returns:
            None: This method does not return a value but will save files directly to the specified output directory.

        Raises:
            ValueError: If required parameters for fetching panorama IDs are not adequately specified.
            FileNotFoundError: If `path_pid` is specified but the file does not exist.

        Notes:
            This method logs significant events and errors, making it suitable for both interactive usage and automated workflows.
        """
        if self.logger is not None:
            # record the arguments
            self.logger.log_args(
                "MLYDownloader download_sv",
                dir_output,
                path_pid=path_pid,
                lat=lat,
                lon=lon,
                input_csv_file=input_csv_file,
                input_shp_file=input_shp_file,
                input_place_name=input_place_name,
                buffer=buffer,
                update_pids=update_pids,
                resolution=resolution,
                cropped=cropped,
                batch_size=batch_size,
                start_date=start_date,
                end_date=end_date,
                metadata_only=metadata_only,
                use_cache=use_cache,
                **kwargs,
            )
        # set necessary directories
        self._set_dirs(dir_output)

        # call _get_pids function first if path_pid is None
        if path_pid is None:
            print("Getting pids...")
            path_pid = self.dir_output / "mly_pids.csv"
            if Path(path_pid).exists() & (not update_pids):
                print("update_pids is set to False. So the following csv file will be used: {}".format(path_pid))
            else:
                self._get_pids(
                    path_pid,
                    lat=lat,
                    lon=lon,
                    input_csv_file=input_csv_file,
                    input_shp_file=input_shp_file,
                    input_place_name=input_place_name,
                    buffer=buffer,
                    start_date=start_date,
                    end_date=end_date,
                    use_cache=use_cache,
                    **kwargs,
                )
        else:
            # check if the path_pid exists
            if Path(path_pid).exists():
                print("The following csv file will be used: {}".format(path_pid))
            else:
                self._get_pids(
                    path_pid,
                    lat=lat,
                    lon=lon,
                    input_csv_file=input_csv_file,
                    input_shp_file=input_shp_file,
                    input_place_name=input_place_name,
                    buffer=buffer,
                    **kwargs,
                )

        # create a folder within self.dir_output
        self.panorama_output = self.dir_output / "mly_svi"
        self.panorama_output.mkdir(parents=True, exist_ok=True)

        # get urls
        if path_pid.exists():
            self._get_urls_mly(path_pid, resolution=resolution, additional_fields=additional_fields)

            # stop if metadata_only is True
            if metadata_only:
                print("The metadata has been downloaded")
                return

            # download images
            self._download_images_mly(path_pid, cropped, batch_size, start_date, end_date)
        else:
            print("There is no panorama ID to download within the given input parameters")

        # delete the cache directory
        if self.dir_cache.exists():
            shutil.rmtree(self.dir_cache)
            print("The cache directory has been deleted")



import math
from concurrent.futures import ThreadPoolExecutor, as_completed
from pathlib import Path
from typing import Union

import cv2
import numpy as np
from tqdm import tqdm

from zensvi.utils.log import Logger


def _xyz2lonlat(xyz):
    """Converts 3D Cartesian coordinates (x, y, z) to geographic coordinates (longitude,
    latitude).

    Args:
        xyz (np.ndarray): An array of shape (..., 3) containing 3D Cartesian coordinates.

    Returns:
        np.ndarray: An array of shape (..., 2) containing longitude and latitude coordinates.
    """
    atan2 = np.arctan2
    asin = np.arcsin

    norm = np.linalg.norm(xyz, axis=-1, keepdims=True)
    xyz_norm = xyz / norm
    x = xyz_norm[..., 0:1]
    y = xyz_norm[..., 1:2]
    z = xyz_norm[..., 2:]

    lon = atan2(x, z)
    lat = asin(y)
    lst = [lon, lat]

    out = np.concatenate(lst, axis=-1)
    return out


def _lonlat2XY(lonlat, shape):
    """Converts geographic coordinates (longitude, latitude) to pixel coordinates (X, Y)
    based on an image shape.

    Args:
        lonlat (np.ndarray): An array of shape (..., 2) containing longitude and latitude coordinates.
        shape (tuple): A tuple (height, width) representing the shape of the image.

    Returns:
        np.ndarray: An array of shape (..., 2) containing pixel coordinates.
    """
    X = (lonlat[..., 0:1] / (2 * np.pi) + 0.5) * (shape[1] - 1)
    Y = (lonlat[..., 1:] / (np.pi) + 0.5) * (shape[0] - 1)
    lst = [X, Y]
    out = np.concatenate(lst, axis=-1)

    return out


class ImageTransformer:
    """Transforms images by applying various projections such as fisheye and perspective
    adjustments.

    Args:
        dir_input (Union[str, Path]): Input directory containing images.
        dir_output (Union[str, Path]): Output directory where transformed images will be saved.

    Raises:
        TypeError: If the input or output directories are not specified as string or Path objects.
    """

    def __init__(
        self,
        dir_input: Union[str, Path],
        dir_output: Union[str, Path],
        log_path: Union[str, Path] = None,
    ):
        if isinstance(dir_input, str):
            dir_input = Path(dir_input)
        elif not isinstance(dir_input, Path):
            raise TypeError("dir_input must be a str or Path object.")
        if isinstance(dir_output, str):
            dir_output = Path(dir_output)
        elif not isinstance(dir_output, Path):
            raise TypeError("dir_output must be a str or Path object.")
        self._dir_input = dir_input
        self._dir_output = dir_output
        # initialize the logger
        self.log_path = log_path
        if self.log_path is not None:
            self.logger = Logger(log_path)
        else:
            self.logger = None

    @property
    def dir_input(self):
        """Property for the input directory.

        Returns:
            Path: dir_input
        """
        return self._dir_input

    @dir_input.setter
    def dir_input(self, value):
        if isinstance(value, str):
            value = Path(value)
        elif not isinstance(value, Path):
            raise TypeError("dir_input must be a str or Path object.")
        self._dir_input = value

    @property
    def dir_output(self):
        """Property for the output directory.

        Returns:
            Path: dir_output
        """
        return self._dir_output

    @dir_output.setter
    def dir_output(self, value):
        if isinstance(value, str):
            value = Path(value)
        elif not isinstance(value, Path):
            raise TypeError("dir_output must be a str or Path object.")
        self._dir_output = value

    def perspective(self, img, FOV, THETA, PHI, height, width):
        """Transforms an image to simulate a perspective view from specific angles.

        Args:
            img (np.ndarray): Source image to transform.
            FOV (float): Field of view in degrees.
            THETA (float): Rotation around the vertical axis in degrees.
            PHI (float): Tilt angle in degrees.
            height (int): Height of the output image.
            width (int): Width of the output image.

        Returns:
            np.ndarray: Transformed image.
        """
        f = 0.5 * width * 1 / np.tan(0.5 * FOV / 180.0 * np.pi)
        cx = (width - 1) / 2.0
        cy = (height - 1) / 2.0
        K = np.array(
            [
                [f, 0, cx],
                [0, f, cy],
                [0, 0, 1],
            ],
            np.float32,
        )
        K_inv = np.linalg.inv(K)

        x = np.arange(width)
        y = np.arange(height)
        x, y = np.meshgrid(x, y)
        z = np.ones_like(x)
        xyz = np.concatenate([x[..., None], y[..., None], z[..., None]], axis=-1)
        xyz = xyz @ K_inv.T

        y_axis = np.array([0.0, 1.0, 0.0], np.float32)
        x_axis = np.array([1.0, 0.0, 0.0], np.float32)
        R1, _ = cv2.Rodrigues(y_axis * np.radians(THETA))
        R2, _ = cv2.Rodrigues(np.dot(R1, x_axis) * np.radians(PHI))
        R = R2 @ R1
        xyz = xyz @ R.T
        lonlat = _xyz2lonlat(xyz)
        XY = _lonlat2XY(lonlat, shape=img.shape).astype(np.float32)
        persp = cv2.remap(img, XY[..., 0], XY[..., 1], cv2.INTER_LINEAR, borderMode=cv2.BORDER_WRAP)
        return persp

    def equidistant_fisheye(self, img):
        """Transforms an image to an equidistant fisheye projection.

        Args:
            img (np.ndarray): Source image to transform.

        Returns:
            np.ndarray: Fisheye projected image.
        """
        # Implem
        rows, cols, c = img.shape
        R = cols / (2 * math.pi)
        D = int(2 * R)
        cx, cy = R, R

        x, y = np.meshgrid(np.arange(D), np.arange(D))
        r = np.sqrt((x - cx) ** 2 + (y - cy) ** 2)
        theta = np.arctan2(y - cy, x - cx)

        xp = np.floor((theta + np.pi) * cols / (2 * np.pi)).astype(int)
        yp = np.floor(r * rows / R).astype(int)

        mask = r < R

        new_img = np.zeros((D, D, c), dtype=np.uint8)
        new_img.fill(255)
        new_img[y[mask], x[mask]] = img[yp[mask], xp[mask]]

        return new_img

    def orthographic_fisheye(self, img):
        """Transforms an image to an orthographic fisheye projection.

        Args:
            img (np.ndarray): Source image to transform.

        Returns:
            np.ndarray: Fisheye projected image.
        """
        rows, cols, c = img.shape
        R = cols / (2 * math.pi)
        D = int(2 * R)
        cx, cy = R, R

        x, y = np.meshgrid(np.arange(D), np.arange(D))
        r = np.sqrt((x - cx) ** 2 + (y - cy) ** 2)
        theta = np.arctan2(y - cy, x - cx)

        xp = np.floor((theta + np.pi) * cols / (2 * np.pi)).astype(int)
        yp = np.floor((2 / np.pi) * np.arcsin(r / R) * rows).astype(int)

        mask = r < R

        new_img = np.zeros((D, D, c), dtype=np.uint8)
        new_img.fill(255)
        new_img[y[mask], x[mask]] = img[yp[mask], xp[mask]]

        return new_img

    def stereographic_fisheye(self, img):
        """Transforms an image to a stereographic fisheye projection.

        Args:
            img (np.ndarray): Source image to transform.

        Returns:
            np.ndarray: Fisheye projected image.
        """
        rows, cols, c = img.shape
        R = cols / (2 * math.pi)
        D = int(2 * R)
        cx, cy = R, R

        # Create a meshgrid of coordinates
        x, y = np.meshgrid(np.arange(D), np.arange(D))
        r = np.sqrt((x - cx) ** 2 + (y - cy) ** 2)
        theta = np.arctan2(y - cy, x - cx)

        # Calculate the new positions in the source image
        xp = np.floor((theta + np.pi) * cols / (2 * np.pi)).astype(np.int32)
        yp = np.floor(rows * np.sin(r / R) / np.sin(1)).astype(np.int32)

        # Clip the values of yp and xp to be within the valid range
        yp = np.clip(yp, 0, rows - 1)
        xp = np.clip(xp, 0, cols - 1)

        # Create a mask for the valid fisheye region
        mask = r <= R

        # Initialize a new image and fill it with a white background
        new_img = np.full((D, D, c), 255, dtype=np.uint8)

        # Use advanced indexing to map the original pixels to the new image
        new_img[y[mask], x[mask]] = img[yp[mask], xp[mask]]

        return new_img

    def equisolid_fisheye(self, img):
        """Transforms an image to an equisolid fisheye projection.

        Args:
            img (np.ndarray): Source image to transform.

        Returns:
            np.ndarray: Fisheye projected image.
        """
        rows, cols, c = img.shape
        R = cols / (2 * math.pi)
        D = int(2 * R)
        cx, cy = R, R

        x, y = np.meshgrid(np.arange(D), np.arange(D))
        r = np.sqrt((x - cx) ** 2 + (y - cy) ** 2)
        theta = np.arctan2(y - cy, x - cx)

        xp = np.floor((theta + np.pi) * cols / (2 * np.pi)).astype(int)
        yp = np.floor(2 * np.sin(r / (2 * R)) * rows).astype(int)

        mask = r < R

        new_img = np.zeros((D, D, c), dtype=np.uint8)
        new_img.fill(255)
        new_img[y[mask], x[mask]] = img[yp[mask], xp[mask]]

        return new_img

    def transform_images(
        self,
        style_list: str = "perspective equidistant_fisheye orthographic_fisheye stereographic_fisheye equisolid_fisheye",
        FOV: Union[int, float] = 90,
        theta: Union[int, float] = 90,
        phi: Union[int, float] = 0,
        aspects: tuple = (9, 16),
        show_size: Union[int, float] = 100,
        use_upper_half: bool = False,
    ):
        """Applies specified transformations to all images in the input directory and
        saves them in the output directory.

        Args:
            style_list (str): Space-separated list of transformation styles to apply. Valid styles include 'perspective',
                            'equidistant_fisheye', 'orthographic_fisheye', 'stereographic_fisheye', and 'equisolid_fisheye'.
            FOV (Union[int, float], optional): Field of view for the 'perspective' style in degrees.
            theta (Union[int, float], optional): Rotation step for generating multiple perspective images in degrees.
            phi (Union[int, float], optional): Tilt angle for the 'perspective' style in degrees.
            aspects (tuple, optional): Aspect ratio of the output images represented as a tuple.
            show_size (Union[int, float], optional): Base size to calculate the dimensions of the output images.
            use_upper_half (bool, optional): If True, only the upper half of the image is used for fisheye transformations.

        Raises:
            ValueError: If an invalid style is specified in style_list.

        Notes:
            This method processes images concurrently, leveraging multi-threading to speed up the transformation tasks. It
            automatically splits style_list into individual styles and processes each style, creating appropriate subdirectories
            in the output directory for each style.
        """
        if self.logger is not None:
            # record the arguments
            self.logger.log_args(
                "transform_images",
                style_list=style_list,
                FOV=FOV,
                theta=theta,
                phi=phi,
                aspects=aspects,
                show_size=show_size,
                use_upper_half=use_upper_half,
            )
        # raise an error if the style_list is a list
        if isinstance(style_list, list):
            raise ValueError("Please input the correct image style as a string, not a list.")
        # check if there's anything other than "perspective" and "fisheye"
        style_list = style_list.split()
        if not all(
            style
            in [
                "perspective",
                "equidistant_fisheye",
                "orthographic_fisheye",
                "stereographic_fisheye",
                "equisolid_fisheye",
            ]
            for style in style_list
        ):
            raise ValueError(
                "Please input the correct image style. The correct image style should be 'perspective', 'equidistant_fisheye', 'orthographic_fisheye', 'stereographic_fisheye', or 'equisolid_fisheye'"
            )

        # set image file extensions
        image_extensions = [
            ".jpg",
            ".jpeg",
            ".png",
            ".bmp",
            ".gif",
            ".tiff",
            ".tif",
            ".webp",
            ".ico",
            ".jfif",
            ".heic",
            ".heif",
        ]

        def run(path_input, path_output, show_size, style, theta, aspects, FOV):
            img_raw = cv2.imread(str(path_input), cv2.IMREAD_COLOR)
            if use_upper_half:
                img_raw = img_raw[: img_raw.shape[0] // 2, :]
            if style == "equidistant_fisheye":
                if not path_output.exists():
                    img_new = self.equidistant_fisheye(img_raw)
                    cv2.imwrite(str(path_output), img_new)

            elif style == "orthographic_fisheye":
                if not path_output.exists():
                    img_new = self.orthographic_fisheye(img_raw)
                    cv2.imwrite(str(path_output), img_new)

            elif style == "stereographic_fisheye":
                if not path_output.exists():
                    img_new = self.stereographic_fisheye(img_raw)
                    cv2.imwrite(str(path_output), img_new)

            elif style == "equisolid_fisheye":
                if not path_output.exists():
                    img_new = self.equisolid_fisheye(img_raw)
                    cv2.imwrite(str(path_output), img_new)

            elif style == "perspective":
                num_images = 360 // theta  # Calculate the number of images based on theta
                thetas = [theta * i for i in range(num_images)]  # Calculate thetas based on step size

                for theta in thetas:
                    height = int(aspects[0] * show_size)
                    width = int(aspects[1] * show_size)
                    aspect_name = "%s--%s" % (aspects[0], aspects[1])
                    path_output_raw = path_output.with_name(
                        f"{path_output.stem}_Direction_{theta}_FOV_{FOV}_aspect_{aspect_name}_raw.png"
                    )
                    if not path_output_raw.exists():
                        img_new = self.perspective(img_raw, FOV, theta, phi, height, width)
                        cv2.imwrite(str(path_output_raw), img_new)

        def process_image(dir_input, dir_output, file_path, show_size, style, theta, aspects, FOV):
            relative_path = file_path.relative_to(dir_input)
            path_output = dir_output / relative_path.with_suffix(".png")
            path_output.parent.mkdir(parents=True, exist_ok=True)
            return file_path, path_output, show_size, style, theta, aspects, FOV

        # Recursive function to get all image files
        def get_image_files(directory):
            for item in directory.iterdir():
                if item.is_file() and item.suffix.lower() in image_extensions:
                    yield item
                elif item.is_dir():
                    yield from get_image_files(item)

        # Check if self.dir_input is a directory or a single file
        if not self.dir_input.is_dir():
            if isinstance(self.dir_input, (str, Path)) and self.dir_input.suffix.lower() in image_extensions:
                dir_input = [Path(self.dir_input)]
                self.dir_input = Path(self.dir_input).parent
            else:
                raise ValueError("Please input a valid directory path or image file.")
        else:
            dir_input = list(get_image_files(self.dir_input))

        for current_style in style_list:
            dir_output = Path(self.dir_output) / current_style
            dir_output.mkdir(parents=True, exist_ok=True)

            with ThreadPoolExecutor() as executor:
                futures = [
                    executor.submit(
                        run,
                        *process_image(
                            self.dir_input,
                            dir_output,
                            file_path,
                            show_size,
                            current_style,
                            theta,
                            aspects,
                            FOV,
                        ),
                    )
                    for file_path in dir_input
                ]
                for future in tqdm(
                    as_completed(futures),
                    total=len(futures),
                    desc=f"Converting to {current_style}",
                ):
                    future.result()



# Copyright (c) Meta Platforms, Inc. and affiliates.
# All rights reserved.
#
# This source code is licensed under the license found in the
# LICENSE file in the root directory of this source tree.

import logging

import torch
import torch.nn as nn
import torch.nn.functional as F

# import torch.distributed as dist


logger = logging.getLogger("dinov2")


class KoLeoLoss(nn.Module):
    """Kozachenko-Leonenko entropic loss regularizer from Sablayrolles et al.

    This loss function is designed to encourage the spreading of vectors for similarity search.

    References:
        - Sablayrolles et al. (2018). Spreading vectors for similarity search.

    Args:
        None

    Returns:
        None
    """

    def __init__(self):
        """Initializes the KoLeoLoss module and sets up the pairwise distance metric."""
        super().__init__()
        self.pdist = nn.PairwiseDistance(2, eps=1e-8)

    def pairwise_NNs_inner(self, x):
        """Computes pairwise nearest neighbors for L2-normalized vectors.

        This method uses Torch instead of Faiss to remain on the GPU.

        Args:
            x (torch.Tensor): Input tensor of shape (B, D) where B is the batch size and D is the feature dimension.

        Returns:
            torch.Tensor: Indices of the nearest neighbors for each vector in the input tensor.
        """
        # parwise dot products (= inverse distance)
        dots = torch.mm(x, x.t())
        n = x.shape[0]
        dots.view(-1)[:: (n + 1)].fill_(-1)  # Trick to fill diagonal with -1
        # max inner prod -> min distance
        _, I = torch.max(dots, dim=1)  # noqa: E741
        return I

    def forward(self, student_output, eps=1e-8):
        """Computes the loss based on the student output.

        Args:
            student_output (torch.Tensor): Backbone output of the student of shape (B, D).
            eps (float, optional): A small value to avoid division by zero. Default is 1e-8.

        Returns:
            torch.Tensor: The computed loss value.
        """
        with torch.cuda.amp.autocast(enabled=False):
            student_output = F.normalize(student_output, eps=eps, p=2, dim=-1)
            I = self.pairwise_NNs_inner(student_output)  # noqa: E741
            distances = self.pdist(student_output, student_output[I])  # BxD, BxD -> B
            loss = -torch.log(distances + eps).mean()
        return loss



from pathlib import Path
from typing import List, Tuple, Union

import pandas as pd
import torch
import tqdm
from huggingface_hub import hf_hub_download
from PIL import Image
from torch.utils.data import DataLoader, Dataset
from torchvision import transforms

from .base import BaseClassifier
from .utils.global_streetscapes import GlobalStreetScapesClassificationModel, glare_dict2idx


class ImageDataset(Dataset):
    """Dataset class for loading and preprocessing images.

    Args:
        image_files: List of paths to image files.

    Attributes:
        image_files: Filtered list of valid image file paths.
        transform: Composition of image transformations to apply.
    """

    def __init__(self, image_files: List[Path]):
        self.image_files = [
            image_file
            for image_file in image_files
            if image_file.suffix.lower() in [".jpg", ".jpeg", ".png"] and not image_file.name.startswith(".")
        ]

        # Image transformations
        self.transform = transforms.Compose(
            [
                transforms.Resize((224, 224)),
                transforms.ToTensor(),
                transforms.Normalize(mean=[0.485, 0.456, 0.406], std=[0.229, 0.224, 0.225]),  # ImageNet normalization
            ]
        )

    def __len__(self) -> int:
        """Returns the number of images in the dataset.

        Returns:
            Number of images.
        """
        return len(self.image_files)

    def __getitem__(self, idx: int) -> Tuple[str, torch.Tensor]:
        """Gets an image and its file path at the given index.

        Args:
            idx: Index of the image to retrieve.

        Returns:
            Tuple containing the image file path as string and the preprocessed image tensor.
        """
        image_file = self.image_files[idx]
        img = Image.open(image_file)  # Open image directly using PIL
        img = self.transform(img)  # Apply transformations

        return str(image_file), img

    def collate_fn(self, data: List[Tuple[str, torch.Tensor]]) -> Tuple[List[str], torch.Tensor]:
        """Custom collate function for batching dataset items.

        Args:
            data: List of tuples containing image file paths and transformed image tensors.

        Returns:
            Tuple containing a list of image file paths and a batch of stacked image tensors.
        """
        image_files, images = zip(*data)
        images = torch.stack(images)  # Stack images to create a batch
        return list(image_files), images


class ClassifierGlare(BaseClassifier):
    """A classifier for identifying glare in images.

    This classifier uses a pre-trained model from Hou et al (2024)
    (https://github.com/ualsg/global-streetscapes) to detect glare in street-level imagery.

    Args:
        device: The device to load the model onto. Options are "cpu", "cuda", or "mps".
            If None, tries to use GPU if available, otherwise falls back to CPU.

    Attributes:
        device: The device the model is loaded on.
        model: The loaded classification model.
    """

    def __init__(self, device=None):
        super().__init__(device)
        self.device = self._get_device(device)

        file_name = "glare_inverse/f6b03038-0831-43ec-88d4-e6c7eb5f8539_glare_glare_inverse_checkpoint.ckpt"
        checkpoint_path = hf_hub_download(
            repo_id="pihalf/gss-models",
            filename=file_name,
            local_dir=Path(__file__).parent.parent.parent.parent.parent / "models",
        )

        checkpoint = torch.load(checkpoint_path, map_location=lambda storage, loc: storage, weights_only=False)

        # Extract the number of classes
        num_classes = checkpoint["state_dict"]["model.classifier.5.weight"].shape[0]

        # Now load the model
        self.model = GlobalStreetScapesClassificationModel.load_from_checkpoint(
            checkpoint_path, num_classes=num_classes, weight=None, strict=False
        )
        self.model.eval()
        self.model.to(self.device)

    def _save_results_to_file(self, results, dir_output, file_name, save_format="csv json"):
        """Saves classification results to file(s) in specified format(s).

        Args:
            results: List of dictionaries containing classification results.
            dir_output: Directory to save output files.
            file_name: Base name for output files (without extension).
            save_format: Space-separated string of formats to save in.
                Options are "csv" and "json". Defaults to "csv json".
        """
        df = pd.DataFrame(results)
        dir_output = Path(dir_output)
        dir_output.mkdir(parents=True, exist_ok=True)
        if "csv" in save_format:
            file_path = dir_output / f"{file_name}.csv"
            df.to_csv(file_path, index=False)
        if "json" in save_format:
            file_path = dir_output / f"{file_name}.json"
            df.to_json(file_path, orient="records")

    def classify(
        self,
        dir_input: Union[str, Path],
        dir_summary_output: Union[str, Path],
        batch_size=1,
        save_format="json csv",
    ) -> List[str]:
        """Classifies images based on presence of glare.

        Processes images from the input directory and classifies them as having glare ("True")
        or not having glare ("False"). Results can be saved in JSON and/or CSV format.

        Args:
            dir_input: Directory containing input images or path to a single image.
            dir_summary_output: Directory to save classification results.
            batch_size: Number of images to process simultaneously. Defaults to 1.
            save_format: Space-separated string of output formats.
                Options are "json" and "csv". Defaults to "json csv".

        Returns:
            List of glare classifications ("True" or "False") for each image.
        """
        # Prepare output directories
        if dir_summary_output:
            Path(dir_summary_output).mkdir(parents=True, exist_ok=True)

        # get all the images in dir_input
        if Path(dir_input).is_file():
            img_paths = [Path(dir_input)]
        else:
            img_paths = [
                p
                for ext in [
                    "*.jpg",
                    "*.jpeg",
                    "*.png",
                    "*.gif",
                    "*.bmp",
                    "*.tiff",
                    "*.JPG",
                    "*.JPEG",
                    "*.PNG",
                    "*.GIF",
                    "*.BMP",
                    "*.TIFF",
                ]
                for p in Path(dir_input).rglob(ext)
            ]

        dataset = ImageDataset(img_paths)
        dataloader = DataLoader(dataset, batch_size=batch_size, shuffle=False, collate_fn=dataset.collate_fn)

        results = []
        # Using torch.no_grad() to avoid unnecessary gradient computations during inference
        with torch.no_grad():
            results = [
                {
                    "filename_key": str(Path(image_file).stem),
                    "glare": glare_dict2idx["index2label"][pred.item()],
                }
                for image_files, images in tqdm.tqdm(dataloader, desc="Classifying glare")
                for image_file, pred in zip(
                    image_files,
                    torch.max(self.model(images.to(self.device, dtype=torch.float32)), 1)[1],
                )
            ]

        # save the results to json and csv
        self._save_results_to_file(
            results,
            dir_summary_output,
            "results",
            save_format=save_format,
        )



import osmnx as ox


def standardize_column_names(df):
    """Standardizes column names for latitude and longitude in a DataFrame.

    Converts various common column names for latitude and longitude coordinates
    to standardized 'latitude' and 'longitude' names.

    Args:
        df (pandas.DataFrame): DataFrame containing coordinate columns.

    Returns:
        pandas.DataFrame: DataFrame with standardized column names.
    """
    longitude_variants = ["longitude", "long", "lon", "lng", "x"]
    latitude_variants = ["latitude", "lat", "lt", "y"]
    # convert all column names to lowercase
    df.columns = [col.lower() for col in df.columns]
    for col in df.columns:
        if col in longitude_variants:
            df.rename(columns={col: "longitude"}, inplace=True)
        elif col in latitude_variants:
            df.rename(columns={col: "latitude"}, inplace=True)
    return df


def create_buffer_gdf(gdf, buffer_distance):
    """Creates a buffer around geometries in a GeoDataFrame.

    Projects the GeoDataFrame if needed, creates a buffer of specified distance
    around the geometries, and reprojects back to WGS84.

    Args:
        gdf (geopandas.GeoDataFrame): GeoDataFrame containing geometries.
        buffer_distance (float): Buffer distance in meters.

    Returns:
        geopandas.GeoDataFrame: GeoDataFrame with buffered geometries.
    """
    if not gdf.crs:
        gdf = gdf.set_crs("EPSG:4326")

    if not gdf.crs.is_projected:
        gdf = ox.projection.project_gdf(gdf)

    # Buffer the points by buffer_distance
    gdf["geometry"] = gdf.buffer(buffer_distance)

    # Project the GeoDataFrame back to EPSG:4326
    gdf = gdf.to_crs("EPSG:4326")

    return gdf


def check_and_buffer(gdf, buffer):
    """Checks geometry type and creates buffer if appropriate.

    Validates that point and line geometries have non-zero buffers,
    then creates the buffer if validation passes.

    Args:
        gdf (geopandas.GeoDataFrame): GeoDataFrame to check and buffer.
        buffer (float): Buffer distance in meters.

    Returns:
        geopandas.GeoDataFrame: Buffered GeoDataFrame if validation passes.

    Raises:
        ValueError: If buffer is 0 for point or line geometries.
    """
    # check geometry type
    geom_type = gdf.geom_type.unique()[0]
    # raise an error if the geometry is a Point/MultiPoint or a LineString/MultiLineString & buffer = 0
    if (
        geom_type == "Point" or geom_type == "MultiPoint" or geom_type == "LineString" or geom_type == "MultiLineString"
    ) and buffer == 0:
        raise ValueError(
            "Buffer cannot be 0 if the geometry is either a Point/MultiPoint or a LineString/MultiLineString"
        )
    else:
        gdf = create_buffer_gdf(gdf, buffer)
        return gdf



# Copyright (c) Meta Platforms, Inc. and affiliates.
# All rights reserved.
#
# This source code is licensed under the license found in the
# LICENSE file in the root directory of this source tree.

import argparse
import gc
import logging
import sys
import time
from typing import List, Optional

import torch
import torch.backends.cudnn as cudnn
import torch.distributed
from cuml.linear_model import LogisticRegression
from dinov2.data import make_dataset
from dinov2.data.transforms import make_classification_eval_transform
from dinov2.distributed import get_global_rank, get_global_size
from dinov2.eval.metrics import MetricType, build_metric
from dinov2.eval.setup import get_args_parser as get_setup_args_parser
from dinov2.eval.setup import setup_and_build_model
from dinov2.eval.utils import evaluate, extract_features
from dinov2.utils.dtype import as_torch_dtype
from torch import nn
from torch.utils.data import TensorDataset
from torchmetrics import MetricTracker

logger = logging.getLogger("dinov2")

DEFAULT_MAX_ITER = 1_000
C_POWER_RANGE = torch.linspace(-6, 5, 45)
_CPU_DEVICE = torch.device("cpu")


def get_args_parser(
    description: Optional[str] = None,
    parents: Optional[List[argparse.ArgumentParser]] = None,
    add_help: bool = True,
):
    """Creates an argument parser for the script.

    Args:
        description (Optional[str]): Description of the parser.
        parents (Optional[List[argparse.ArgumentParser]]): List of parent parsers.
        add_help (bool): Whether to add a help option. Defaults to True.

    Returns:
        argparse.ArgumentParser: The configured argument parser.
    """
    parents = parents or []
    setup_args_parser = get_setup_args_parser(parents=parents, add_help=False)
    parents = [setup_args_parser]
    parser = argparse.ArgumentParser(
        description=description,
        parents=parents,
        add_help=add_help,
    )
    parser.add_argument(
        "--train-dataset",
        dest="train_dataset_str",
        type=str,
        help="Training dataset",
    )
    parser.add_argument(
        "--val-dataset",
        dest="val_dataset_str",
        type=str,
        help="Validation dataset",
    )
    parser.add_argument(
        "--finetune-dataset-str",
        dest="finetune_dataset_str",
        type=str,
        help="Fine-tuning dataset",
    )
    parser.add_argument(
        "--finetune-on-val",
        action="store_true",
        help="If there is no finetune dataset, whether to choose the "
        "hyperparameters on the val set instead of 10%% of the train dataset",
    )
    parser.add_argument(
        "--metric-type",
        type=MetricType,
        choices=list(MetricType),
        help="Metric type",
    )
    parser.add_argument(
        "--train-features-device",
        type=str,
        help="Device to gather train features (cpu, cuda, cuda:0, etc.), default: %(default)s",
    )
    parser.add_argument(
        "--train-dtype",
        type=str,
        help="Data type to convert the train features to (default: %(default)s)",
    )
    parser.add_argument(
        "--max-train-iters",
        type=int,
        help="Maximum number of train iterations (default: %(default)s)",
    )
    parser.set_defaults(
        train_dataset_str="ImageNet:split=TRAIN",
        val_dataset_str="ImageNet:split=VAL",
        finetune_dataset_str=None,
        metric_type=MetricType.MEAN_ACCURACY,
        train_features_device="cpu",
        train_dtype="float64",
        max_train_iters=DEFAULT_MAX_ITER,
        finetune_on_val=False,
    )
    return parser


class LogRegModule(nn.Module):
    """Logistic Regression Module using cuML's LogisticRegression.

    Attributes:
        dtype (torch.dtype): Data type for the model.
        device (torch.device): Device for the model.
        estimator (LogisticRegression): The cuML logistic regression estimator.
    """

    def __init__(
        self,
        C,
        max_iter=DEFAULT_MAX_ITER,
        dtype=torch.float64,
        device=_CPU_DEVICE,
    ):
        """Initializes the LogRegModule.

        Args:
            C (float): Inverse of regularization strength.
            max_iter (int): Maximum number of iterations for optimization.
            dtype (torch.dtype): Data type for the model. Defaults to torch.float64.
            device (torch.device): Device for the model. Defaults to CPU.
        """
        super().__init__()
        self.dtype = dtype
        self.device = device
        self.estimator = LogisticRegression(
            penalty="l2",
            C=C,
            max_iter=max_iter,
            output_type="numpy",
            tol=1e-12,
            linesearch_max_iter=50,
        )

    def forward(self, samples, targets):
        """Forward pass for the model.

        Args:
            samples (torch.Tensor): Input samples.
            targets (torch.Tensor): Target labels.

        Returns:
            dict: A dictionary containing predictions and targets.
        """
        samples_device = samples.device
        samples = samples.to(dtype=self.dtype, device=self.device)
        if self.device == _CPU_DEVICE:
            samples = samples.numpy()
        probas = self.estimator.predict_proba(samples)
        return {"preds": torch.from_numpy(probas).to(samples_device), "target": targets}

    def fit(self, train_features, train_labels):
        """Fits the logistic regression model to the training data.

        Args:
            train_features (torch.Tensor): Features for training.
            train_labels (torch.Tensor): Labels for training.
        """
        train_features = train_features.to(dtype=self.dtype, device=self.device)
        train_labels = train_labels.to(dtype=self.dtype, device=self.device)
        if self.device == _CPU_DEVICE:
            # both cuML and sklearn only work with numpy arrays on CPU
            train_features = train_features.numpy()
            train_labels = train_labels.numpy()
        self.estimator.fit(train_features, train_labels)


def evaluate_model(*, logreg_model, logreg_metric, test_data_loader, device):
    """Evaluates the logistic regression model.

    Args:
        logreg_model (LogRegModule): The logistic regression model.
        logreg_metric: Metric tracker for evaluation.
        test_data_loader: DataLoader for the test dataset.
        device (torch.device): Device for evaluation.

    Returns:
        Evaluation results.
    """
    postprocessors = {"metrics": logreg_model}
    metrics = {"metrics": logreg_metric}
    return evaluate(nn.Identity(), test_data_loader, postprocessors, metrics, device)


def train_for_C(
    *,
    C,
    max_iter,
    train_features,
    train_labels,
    dtype=torch.float64,
    device=_CPU_DEVICE,
):
    """Trains the logistic regression model for a specific C value.

    Args:
        C (float): Inverse of regularization strength.
        max_iter (int): Maximum number of iterations for optimization.
        train_features (torch.Tensor): Features for training.
        train_labels (torch.Tensor): Labels for training.
        dtype (torch.dtype): Data type for the model. Defaults to torch.float64.
        device (torch.device): Device for the model. Defaults to CPU.

    Returns:
        LogRegModule: The trained logistic regression model.
    """
    logreg_model = LogRegModule(C, max_iter=max_iter, dtype=dtype, device=device)
    logreg_model.fit(train_features, train_labels)
    return logreg_model


def train_and_evaluate(
    *,
    C,
    max_iter,
    train_features,
    train_labels,
    logreg_metric,
    test_data_loader,
    train_dtype=torch.float64,
    train_features_device,
    eval_device,
):
    """Trains and evaluates the logistic regression model.

    Args:
        C (float): Inverse of regularization strength.
        max_iter (int): Maximum number of iterations for optimization.
        train_features (torch.Tensor): Features for training.
        train_labels (torch.Tensor): Labels for training.
        logreg_metric: Metric tracker for evaluation.
        test_data_loader: DataLoader for the test dataset.
        train_dtype (torch.dtype): Data type for the model. Defaults to torch.float64.
        train_features_device (torch.device): Device for training features.
        eval_device (torch.device): Device for evaluation.

    Returns:
        Evaluation results.
    """
    logreg_model = train_for_C(
        C=C,
        max_iter=max_iter,
        train_features=train_features,
        train_labels=train_labels,
        dtype=train_dtype,
        device=train_features_device,
    )
    return evaluate_model(
        logreg_model=logreg_model,
        logreg_metric=logreg_metric,
        test_data_loader=test_data_loader,
        device=eval_device,
    )


def sweep_C_values(
    *,
    train_features,
    train_labels,
    test_data_loader,
    metric_type,
    num_classes,
    train_dtype=torch.float64,
    train_features_device=_CPU_DEVICE,
    max_train_iters=DEFAULT_MAX_ITER,
):
    """Sweeps over a range of C values to find the best one.

    Args:
        train_features (torch.Tensor): Features for training.
        train_labels (torch.Tensor): Labels for training.
        test_data_loader: DataLoader for the test dataset.
        metric_type: Type of metric for evaluation.
        num_classes (int): Number of classes in the dataset.
        train_dtype (torch.dtype): Data type for the model. Defaults to torch.float64.
        train_features_device (torch.device): Device for training features. Defaults to CPU.
        max_train_iters (int): Maximum number of iterations for training. Defaults to DEFAULT_MAX_ITER.

    Returns:
        tuple: Best statistics and best C value.
    """
    if metric_type == MetricType.PER_CLASS_ACCURACY:
        # If we want to output per-class accuracy, we select the hyperparameters with mean per class
        metric_type = MetricType.MEAN_PER_CLASS_ACCURACY
    logreg_metric = build_metric(metric_type, num_classes=num_classes)
    metric_tracker = MetricTracker(logreg_metric, maximize=True)
    ALL_C = 10**C_POWER_RANGE
    logreg_models = {}

    train_features = train_features.to(dtype=train_dtype, device=train_features_device)
    train_labels = train_labels.to(device=train_features_device)

    for i in range(get_global_rank(), len(ALL_C), get_global_size()):
        C = ALL_C[i].item()
        logger.info(
            f"Training for C = {C:.5f}, dtype={train_dtype}, "
            f"features: {train_features.shape}, {train_features.dtype}, "
            f"labels: {train_labels.shape}, {train_labels.dtype}"
        )
        logreg_models[C] = train_for_C(
            C=C,
            max_iter=max_train_iters,
            train_features=train_features,
            train_labels=train_labels,
            dtype=train_dtype,
            device=train_features_device,
        )

    gather_list = [None for _ in range(get_global_size())]
    torch.distributed.all_gather_object(gather_list, logreg_models)

    logreg_models_gathered = {}
    for logreg_dict in gather_list:
        logreg_models_gathered.update(logreg_dict)

    for i in range(len(ALL_C)):
        metric_tracker.increment()
        C = ALL_C[i].item()
        evals = evaluate_model(
            logreg_model=logreg_models_gathered[C],
            logreg_metric=metric_tracker,
            test_data_loader=test_data_loader,
            device=torch.cuda.current_device(),
        )
        logger.info(f"Trained for C = {C:.5f}, accuracies = {evals}")

        best_stats, which_epoch = metric_tracker.best_metric(return_step=True)
        best_stats_100 = {k: 100.0 * v for k, v in best_stats.items()}
        if which_epoch["top-1"] == i:
            best_C = C
    logger.info(f"Sweep best {best_stats_100}, best C = {best_C:.6f}")

    return best_stats, best_C


def eval_log_regression(
    *,
    model,
    train_dataset,
    val_dataset,
    finetune_dataset,
    metric_type,
    batch_size,
    num_workers,
    finetune_on_val=False,
    train_dtype=torch.float64,
    train_features_device=_CPU_DEVICE,
    max_train_iters=DEFAULT_MAX_ITER,
):
    """Implements the "standard" process for log regression evaluation.

    The value of C is chosen by training on train_dataset and evaluating on
    finetune_dataset. Then, the final model is trained on a concatenation of
    train_dataset and finetune_dataset, and is evaluated on val_dataset. If there is no
    finetune_dataset, the value of C is the one that yields the best results on a random
    10% subset of the train dataset.

    Args:
        model: The model to extract features from.
        train_dataset: Dataset for training.
        val_dataset: Dataset for validation.
        finetune_dataset: Dataset for fine-tuning.
        metric_type: Type of metric for evaluation.
        batch_size (int): Batch size for data loading.
        num_workers (int): Number of workers for data loading.
        finetune_on_val (bool): Whether to fine-tune on validation dataset. Defaults to False.
        train_dtype (torch.dtype): Data type for the model. Defaults to torch.float64.
        train_features_device (torch.device): Device for training features. Defaults to CPU.
        max_train_iters (int): Maximum number of iterations for training. Defaults to DEFAULT_MAX_ITER.

    Returns:
        dict: Best statistics from the evaluation.
    """
    start = time.time()

    train_features, train_labels = extract_features(
        model,
        train_dataset,
        batch_size,
        num_workers,
        gather_on_cpu=(train_features_device == _CPU_DEVICE),
    )
    val_features, val_labels = extract_features(
        model,
        val_dataset,
        batch_size,
        num_workers,
        gather_on_cpu=(train_features_device == _CPU_DEVICE),
    )
    val_data_loader = torch.utils.data.DataLoader(
        TensorDataset(val_features, val_labels),
        batch_size=batch_size,
        drop_last=False,
        num_workers=0,
        persistent_workers=False,
    )

    if finetune_dataset is None and finetune_on_val:
        logger.info("Choosing hyperparameters on the val dataset")
        finetune_features, finetune_labels = val_features, val_labels
    elif finetune_dataset is None and not finetune_on_val:
        logger.info("Choosing hyperparameters on 10% of the train dataset")
        torch.manual_seed(0)
        indices = torch.randperm(len(train_features), device=train_features.device)
        finetune_index = indices[: len(train_features) // 10]
        train_index = indices[len(train_features) // 10 :]
        finetune_features, finetune_labels = (
            train_features[finetune_index],
            train_labels[finetune_index],
        )
        train_features, train_labels = (
            train_features[train_index],
            train_labels[train_index],
        )
    else:
        logger.info("Choosing hyperparameters on the finetune dataset")
        finetune_features, finetune_labels = extract_features(
            model,
            finetune_dataset,
            batch_size,
            num_workers,
            gather_on_cpu=(train_features_device == _CPU_DEVICE),
        )
    # release the model - free GPU memory
    del model
    gc.collect()
    torch.cuda.empty_cache()
    finetune_data_loader = torch.utils.data.DataLoader(
        TensorDataset(finetune_features, finetune_labels),
        batch_size=batch_size,
        drop_last=False,
    )

    if len(train_labels.shape) > 1:
        num_classes = train_labels.shape[1]
    else:
        num_classes = train_labels.max() + 1

    logger.info("Using cuML for logistic regression")

    best_stats, best_C = sweep_C_values(
        train_features=train_features,
        train_labels=train_labels,
        test_data_loader=finetune_data_loader,
        metric_type=metric_type,
        num_classes=num_classes,
        train_dtype=train_dtype,
        train_features_device=train_features_device,
        max_train_iters=max_train_iters,
    )

    if not finetune_on_val:
        logger.info("Best parameter found, concatenating features")
        train_features = torch.cat((train_features, finetune_features))
        train_labels = torch.cat((train_labels, finetune_labels))

    logger.info("Training final model")
    logreg_metric = build_metric(metric_type, num_classes=num_classes)
    evals = train_and_evaluate(
        C=best_C,
        max_iter=max_train_iters,
        train_features=train_features,
        train_labels=train_labels,
        logreg_metric=logreg_metric.clone(),
        test_data_loader=val_data_loader,
        eval_device=torch.cuda.current_device(),
        train_dtype=train_dtype,
        train_features_device=train_features_device,
    )

    best_stats = evals[1]["metrics"]

    best_stats["best_C"] = best_C

    logger.info(f"Log regression evaluation done in {int(time.time() - start)}s")
    return best_stats


def eval_log_regression_with_model(
    model,
    train_dataset_str="ImageNet:split=TRAIN",
    val_dataset_str="ImageNet:split=VAL",
    finetune_dataset_str=None,
    autocast_dtype=torch.float,
    finetune_on_val=False,
    metric_type=MetricType.MEAN_ACCURACY,
    train_dtype=torch.float64,
    train_features_device=_CPU_DEVICE,
    max_train_iters=DEFAULT_MAX_ITER,
):
    """Evaluates the logistic regression model with the specified datasets.

    Args:
        model: The model to extract features from.
        train_dataset_str (str): String representation of the training dataset. Defaults to "ImageNet:split=TRAIN".
        val_dataset_str (str): String representation of the validation dataset. Defaults to "ImageNet:split=VAL".
        finetune_dataset_str (str): String representation of the fine-tuning dataset. Defaults to None.
        autocast_dtype (torch.dtype): Data type for automatic mixed precision. Defaults to torch.float.
        finetune_on_val (bool): Whether to fine-tune on validation dataset. Defaults to False.
        metric_type: Type of metric for evaluation. Defaults to MetricType.MEAN_ACCURACY.
        train_dtype (torch.dtype): Data type for the model. Defaults to torch.float64.
        train_features_device (torch.device): Device for training features. Defaults to CPU.
        max_train_iters (int): Maximum number of iterations for training. Defaults to DEFAULT_MAX_ITER.

    Returns:
        dict: Results of the evaluation including accuracies and best C value.
    """
    cudnn.benchmark = True

    transform = make_classification_eval_transform(resize_size=224)
    target_transform = None

    train_dataset = make_dataset(
        dataset_str=train_dataset_str,
        transform=transform,
        target_transform=target_transform,
    )
    val_dataset = make_dataset(
        dataset_str=val_dataset_str,
        transform=transform,
        target_transform=target_transform,
    )
    if finetune_dataset_str is not None:
        finetune_dataset = make_dataset(
            dataset_str=finetune_dataset_str,
            transform=transform,
            target_transform=target_transform,
        )
    else:
        finetune_dataset = None

    with torch.cuda.amp.autocast(dtype=autocast_dtype):
        results_dict_logreg = eval_log_regression(
            model=model,
            train_dataset=train_dataset,
            val_dataset=val_dataset,
            finetune_dataset=finetune_dataset,
            metric_type=metric_type,
            batch_size=256,
            num_workers=0,  # 5,
            finetune_on_val=finetune_on_val,
            train_dtype=train_dtype,
            train_features_device=train_features_device,
            max_train_iters=max_train_iters,
        )

    results_dict = {
        "top-1": results_dict_logreg["top-1"].cpu().numpy() * 100.0,
        "top-5": results_dict_logreg.get("top-5", torch.tensor(0.0)).cpu().numpy() * 100.0,
        "best_C": results_dict_logreg["best_C"],
    }
    logger.info(
        "\n".join(
            [
                "Training of the supervised logistic regression on frozen features completed.\n"
                "Top-1 test accuracy: {acc:.1f}".format(acc=results_dict["top-1"]),
                "Top-5 test accuracy: {acc:.1f}".format(acc=results_dict["top-5"]),
                "obtained for C = {c:.6f}".format(c=results_dict["best_C"]),
            ]
        )
    )

    torch.distributed.barrier()
    return results_dict


def main(args):
    """Main function to execute the logistic regression evaluation.

    Args:
        args: Command line arguments.

    Returns:
        int: Exit status code.
    """
    model, autocast_dtype = setup_and_build_model(args)
    eval_log_regression_with_model(
        model=model,
        train_dataset_str=args.train_dataset_str,
        val_dataset_str=args.val_dataset_str,
        finetune_dataset_str=args.finetune_dataset_str,
        autocast_dtype=autocast_dtype,
        finetune_on_val=args.finetune_on_val,
        metric_type=args.metric_type,
        train_dtype=as_torch_dtype(args.train_dtype),
        train_features_device=torch.device(args.train_features_device),
        max_train_iters=args.max_train_iters,
    )
    return 0


if __name__ == "__main__":
    description = "DINOv2 logistic regression evaluation"
    args_parser = get_args_parser(description=description)
    args = args_parser.parse_args()
    sys.exit(main(args))



from concurrent.futures import ProcessPoolExecutor, ThreadPoolExecutor, as_completed

import geopandas as gpd
import networkx
import numpy as np
import osmnx as ox
import pandas as pd
from pyproj import Transformer
from shapely.geometry import Point
from tqdm.auto import tqdm

tqdm.pandas()


class GeoProcessor:
    """A class for processing geographic data and extracting points.

    This class handles various geometry types (Point, LineString, Polygon, etc.) and can
    extract points either by sampling along features or creating regular grids.

    Args:
        gdf (geopandas.GeoDataFrame): Input geodataframe containing geometries to process
        distance (float, optional): Sampling distance in meters. Defaults to 1.
        grid (bool, optional): Whether to use grid-based sampling. Defaults to False.
        grid_size (float, optional): Grid cell size in meters. Defaults to 1.
        id_columns (list, optional): List of columns to preserve in output. Defaults to [].
        **kwargs: Additional keyword arguments:
            network_type (str): OSM network type to use. Defaults to "all".
            custom_filter (str): Custom OSM filter query.
    """

    def __init__(self, gdf, distance=1, grid=False, grid_size=1, id_columns=[], **kwargs):
        self.gdf = gdf
        self.distance = distance
        self.processing_functions = {
            "Point": self.process_point,
            "MultiPoint": self.process_multipoint,
            "LineString": self.process_linestring,
            "MultiLineString": self.process_multilinestring,
            "Polygon": self.process_polygon,
            "MultiPolygon": self.process_multipolygon,
        }
        self.grid = grid
        self.grid_size = grid_size
        self.utm_crs = None
        self.id_columns = id_columns
        if "network_type" in kwargs:
            self.network_type = kwargs["network_type"]
        else:
            self.network_type = "all"

        if "custom_filter" in kwargs:
            self.custom_filter = kwargs["custom_filter"]
        else:
            self.custom_filter = None

    def get_lat_lon(self):
        """Extract latitude and longitude points from input geometries.

        Processes each geometry type using the appropriate method and combines results.

        Returns:
            pandas.DataFrame: DataFrame containing extracted points with longitude and latitude columns.
        """
        self.gdf["feature_type"] = self.gdf["geometry"].apply(lambda x: x.geom_type)
        gdf_list = []

        for feature_type, func in self.processing_functions.items():
            sub_gdf = self.gdf[self.gdf["feature_type"] == feature_type]
            if not sub_gdf.empty:
                print(f"Getting longitude and latitude from {feature_type} feature")
                processed_gdf = func(sub_gdf)
                gdf_list.append(processed_gdf)

        result_gdf = pd.concat(gdf_list)
        return result_gdf

    def process_point(self, gdf):
        """Process Point geometries.

        Args:
            gdf (geopandas.GeoDataFrame): GeoDataFrame containing Point geometries.

        Returns:
            pandas.DataFrame: DataFrame with longitude and latitude columns.
        """
        gdf["longitude"] = gdf.geometry.x
        gdf["latitude"] = gdf.geometry.y
        return gdf[self.id_columns + ["longitude", "latitude"]]

    def process_multipoint(self, gdf):
        """Process MultiPoint geometries by exploding into individual points.

        Args:
            gdf (geopandas.GeoDataFrame): GeoDataFrame containing MultiPoint geometries.

        Returns:
            pandas.DataFrame: DataFrame with longitude and latitude columns.
        """
        gdf = gdf.explode("geometry").reset_index(drop=True)
        return self.process_point(gdf)

    def process_linestring(self, gdf):
        """Process LineString geometries by interpolating points along lines.

        Args:
            gdf (geopandas.GeoDataFrame): GeoDataFrame containing LineString geometries.

        Returns:
            pandas.DataFrame: DataFrame with longitude and latitude columns.
        """
        gdf_utm = ox.projection.project_gdf(gdf)
        self.utm_crs = gdf_utm.crs

        gdf_utm["sample_points"] = gdf_utm["geometry"].progress_apply(
            lambda geom: list(ox.utils_geo.interpolate_points(geom, dist=self.distance))
        )
        gdf_utm = gdf_utm.explode("sample_points").reset_index(drop=True)

        gdf_utm["longitude"], gdf_utm["latitude"] = zip(*self.utm_to_lat_lon(gdf_utm["sample_points"], self.utm_crs))
        return gdf_utm[self.id_columns + ["longitude", "latitude"]]

    def process_multilinestring(self, gdf):
        """Process MultiLineString geometries by exploding into individual LineStrings.

        Args:
            gdf (geopandas.GeoDataFrame): GeoDataFrame containing MultiLineString geometries.

        Returns:
            pandas.DataFrame: DataFrame with longitude and latitude columns.
        """
        gdf = gdf.explode("geometry").reset_index(drop=True)
        return self.process_linestring(gdf)

    def get_street_points(self, polygon):
        """Extract points along street network within a polygon.

        Args:
            polygon (shapely.geometry.Polygon): Polygon to extract street network from.

        Returns:
            tuple: List of points and their UTM CRS.
        """
        graph = ox.graph_from_polygon(polygon, network_type=self.network_type, custom_filter=self.custom_filter)
        _, edges = ox.graph_to_gdfs(graph)
        edges_utm = ox.projection.project_gdf(edges)
        utm_crs = edges_utm.crs

        edges_utm["sample_points"] = edges_utm["geometry"].apply(
            lambda geom: list(ox.utils_geo.interpolate_points(geom, dist=self.distance))
        )
        edges_utm = edges_utm.explode("sample_points").reset_index(drop=True)
        return edges_utm["sample_points"].tolist(), utm_crs

    def create_point_grid(self, polygon, grid_size, crs="EPSG:4326"):
        """Create a regular grid of points within a polygon's bounding box.

        Args:
            polygon (geopandas.GeoDataFrame): Polygon to create grid within.
            grid_size (float): Grid cell size in meters.
            crs (str, optional): Coordinate reference system. Defaults to "EPSG:4326".

        Returns:
            tuple: List of points in UTM coordinates and their UTM CRS.
        """
        west, south, east, north = polygon.bounds

        polygon_geom = ox.utils_geo.bbox_to_poly(north, south, east, west)
        bounding_box = gpd.GeoDataFrame({"geometry": [polygon_geom]}, crs=crs)
        bounding_box_utm = ox.projection.project_gdf(bounding_box)
        utm_crs = bounding_box_utm.crs
        west_utm, south_utm, east_utm, north_utm = bounding_box_utm.total_bounds

        x_coords = np.arange(west_utm, east_utm, grid_size)
        y_coords = np.arange(south_utm, north_utm, grid_size)
        x_grid, y_grid = np.meshgrid(x_coords, y_coords)
        points = np.vstack((x_grid.ravel(), y_grid.ravel())).T
        points_utm = gpd.GeoSeries((Point(xy) for xy in points), crs=utm_crs)

        return [(point.x, point.y) for point in points_utm], utm_crs

    def utm_to_lat_lon(self, utm_points, utm_crs):
        """Convert UTM coordinates to latitude/longitude.

        Args:
            utm_points (list): List of points in UTM coordinates.
            utm_crs (str): UTM coordinate reference system.

        Returns:
            list: Points converted to latitude/longitude coordinates.
        """
        transformer = Transformer.from_crs(utm_crs, "EPSG:4326", always_xy=True)
        return [transformer.transform(*point) for point in utm_points]

    def process_polygon(self, gdf):
        """Process Polygon geometries by extracting points from street network or grid.

        Args:
            gdf (geopandas.GeoDataFrame): GeoDataFrame containing Polygon geometries.

        Returns:
            pandas.DataFrame: DataFrame with longitude and latitude columns.
        """
        batch_size = 100
        num_batches = (len(gdf) + batch_size - 1) // batch_size
        failed_geoms = []
        results = []

        for i in tqdm(
            range(num_batches),
            desc=f"Processing polygon by batch size {min(batch_size, len(gdf))}",
        ):
            with ProcessPoolExecutor() as executor:
                batch_futures = {}
                for geom in gdf.geometry.iloc[i * batch_size : (i + 1) * batch_size]:
                    if not self.grid:
                        future = executor.submit(self.get_street_points, geom)
                    else:
                        future = executor.submit(self.create_point_grid, geom, self.grid_size)
                    batch_futures[future] = geom

                for future in tqdm(
                    as_completed(batch_futures.keys()),
                    total=len(batch_futures),
                    desc=f"Processing polygon for batch #{i+1}",
                ):
                    geom = batch_futures[future]
                    try:
                        result = future.result()
                        results.append(result)
                    except (ValueError, networkx.exception.NetworkXPointlessConcept):
                        failed_geoms.append(geom)

        if len(failed_geoms) > 0:
            print("Retrying failed geoms by making grids")
            with ProcessPoolExecutor() as executor:
                retry_futures = {}
                for geom in tqdm(failed_geoms, desc="Preparing Failed Geoms"):
                    future = executor.submit(self.create_point_grid, geom, self.grid_size)
                    retry_futures[future] = geom

                for future in tqdm(
                    as_completed(retry_futures.keys()),
                    total=len(retry_futures),
                    desc="Processing Failed Geoms",
                ):
                    result = future.result()
                    results.append(result)

        gdf["street_points"] = [result[0] for result in results]
        gdf["utm_crs"] = [result[1] for result in results]

        gdf_exploded = gdf.explode("street_points").reset_index(drop=True)
        gdf_exploded["utm_crs"] = gdf_exploded["utm_crs"].astype(str)

        with ThreadPoolExecutor() as executor:
            futures = {
                executor.submit(self.utm_to_lat_lon, group["street_points"], crs): crs
                for crs, group in gdf_exploded.groupby("utm_crs")
            }
            lat_lon_points = []
            for future in tqdm(
                as_completed(futures),
                total=len(futures),
                desc="Converting UTM to Lat/Lon",
            ):
                lat_lon_points.extend(future.result())

        gdf_exploded[["longitude", "latitude"]] = pd.DataFrame(lat_lon_points, index=gdf_exploded.index)

        return gdf_exploded[self.id_columns + ["longitude", "latitude"]]

    def process_multipolygon(self, gdf):
        """Process MultiPolygon geometries by exploding into individual Polygons.

        Args:
            gdf (geopandas.GeoDataFrame): GeoDataFrame containing MultiPolygon geometries.

        Returns:
            pandas.DataFrame: DataFrame with longitude and latitude columns.
        """
        gdf = gdf.explode("geometry").reset_index(drop=True)
        return self.process_polygon(gdf)



# Copyright (c) Meta Platforms, Inc. and affiliates.
# All rights reserved.
#
# This source code is licensed under the license found in the
# LICENSE file in the root directory of this source tree.

import logging
import os
import sys

from dinov2.eval.knn import get_args_parser as get_knn_args_parser
from dinov2.logging import setup_logging
from dinov2.run.submit import get_args_parser, submit_jobs

logger = logging.getLogger("dinov2")


class Evaluator:
    """Evaluator class for managing the k-NN evaluation process.

    Args:
        args: Command line arguments for the evaluation.
    """

    def __init__(self, args):
        """Initializes the Evaluator with the given arguments.

        Args:
            args: Command line arguments for the evaluation.
        """
        self.args = args

    def __call__(self):
        """Executes the evaluation process by setting up arguments and calling the k-NN evaluation main function."""
        from dinov2.eval.knn import main as knn_main

        self._setup_args()
        knn_main(self.args)

    def checkpoint(self):
        """Creates a checkpoint for the evaluation process.

        Returns:
            A delayed submission for the evaluation job.
        """
        import submitit

        logger.info(f"Requeuing {self.args}")
        empty = type(self)(self.args)
        return submitit.helpers.DelayedSubmission(empty)

    def _setup_args(self):
        """Sets up the arguments for the evaluation job, including job environment details."""
        import submitit

        job_env = submitit.JobEnvironment()
        self.args.output_dir = self.args.output_dir.replace("%j", str(job_env.job_id))
        logger.info(f"Process group: {job_env.num_tasks} tasks, rank: {job_env.global_rank}")
        logger.info(f"Args: {self.args}")


def main():
    """Main function to launch the DINOv2 k-NN evaluation.

    Returns:
        int: Exit status code.
    """
    description = "Submitit launcher for DINOv2 k-NN evaluation"
    knn_args_parser = get_knn_args_parser(add_help=False)
    parents = [knn_args_parser]
    args_parser = get_args_parser(description=description, parents=parents)
    args = args_parser.parse_args()

    setup_logging()

    assert os.path.exists(args.config_file), "Configuration file does not exist!"
    submit_jobs(Evaluator, args, name="dinov2:knn")
    return 0


if __name__ == "__main__":
    sys.exit(main())



import json
import os
import random
import warnings
from concurrent.futures import ThreadPoolExecutor, as_completed
from io import BytesIO
from pathlib import Path

import geopandas as gpd
import osmnx as ox
import pandas as pd
import requests
from PIL import Image
from shapely.geometry import Point
from tqdm import tqdm

from zensvi.download.base import BaseDownloader
from zensvi.download.utils.geoprocess import GeoProcessor
from zensvi.download.utils.helpers import standardize_column_names
from zensvi.utils.log import Logger

warnings.filterwarnings("ignore", category=UserWarning)


class AMSDownloader(BaseDownloader):
    """Amsterdam Street View Downloader class.

    Args:
        log_path (str, optional): Path to the log file. Defaults to
            None.
        max_workers (int, optional): Number of workers for parallel processing. Defaults to None.
    """

    def __init__(self, log_path=None, max_workers=None):
        super().__init__(log_path)
        self._max_workers = max_workers
        if log_path is not None:
            self.logger = Logger(log_path)
        else:
            self.logger = None

    @property
    def max_workers(self):
        """Property for the number of workers for parallel processing.

        Returns:
            int: max_workers
        """
        return self._max_workers

    @max_workers.setter
    def max_workers(self, max_workers):
        if max_workers is None:
            self._max_workers = min(32, os.cpu_count() + 4)
        else:
            self._max_workers = max_workers

    def _set_dirs(self, dir_output):
        self.dir_output = Path(dir_output)
        self.dir_output.mkdir(parents=True, exist_ok=True)
        self.dir_cache = self.dir_output / "cache_zensvi"
        self.dir_cache.mkdir(parents=True, exist_ok=True)
        self.cache_lat_lon = self.dir_cache / "lat_lon.csv"
        self.cache_pids_raw = self.dir_cache / "pids_raw.csv"

    def _get_pids_from_df(self, df, id_columns=None):
        if self.cache_lat_lon.exists():
            df = pd.read_csv(self.cache_lat_lon)
            print("The lat and lon have been read from the cache")
        else:
            if isinstance(df, gpd.GeoDataFrame):
                gp = GeoProcessor(
                    df,
                    distance=self.distance,
                    grid=self.grid,
                    grid_size=self.grid_size,
                )
                df = gp.get_lat_lon()
            df["lat_lon_id"] = range(1, len(df) + 1)
            df.to_csv(self.cache_lat_lon, index=False)

        if self.cache_pids_raw.exists():
            print("The raw panorama IDs have been read from the cache")
            results_df = pd.read_csv(self.cache_pids_raw)
        else:
            results = []

            def worker(row):
                pids = self._get_raw_pids(row.latitude, row.longitude, self.buffer)
                result = []
                for pid in pids:
                    result.append(
                        {
                            "lat_lon_id": row.lat_lon_id,
                            "input_latitude": row.latitude,
                            "input_longitude": row.longitude,
                            "pano_id": pid,
                        }
                    )
                return result

            with ThreadPoolExecutor() as executor:
                futures = {executor.submit(worker, row): row for _, row in df.iterrows()}
                for future in tqdm(as_completed(futures), total=len(futures), desc="Getting pids"):
                    try:
                        results.extend(future.result())
                    except Exception as e:
                        print(f"Error: {e}")

            results_df = pd.DataFrame(results)
            # drop duplicates in pano_id
            results_df = results_df.drop_duplicates(subset=["pano_id"])
            results_df.to_csv(self.cache_pids_raw, index=False)

        return results_df

    def _get_raw_pids(self, lat, lon, buffer):
        """Get raw panorama IDs from the Amsterdam Street View API."""
        url = f"https://api.data.amsterdam.nl/panorama/panoramas/?near={lon},{lat}&radius={buffer}&srid=4326"

        for attempt in range(10):
            try:
                proxy = random.choice(self._proxies)
                user_agent = random.choice(self._user_agents)
                headers = {"User-Agent": user_agent["user_agent"]}  # Extract the string from the dictionary
                response = requests.get(url, proxies=proxy, headers=headers)
                response.raise_for_status()
                data = json.loads(response.content)
                panoramas = data["_embedded"]["panoramas"]
                return [item["pano_id"] for item in panoramas]
            except Exception as e:
                if attempt == 9:  # Last attempt
                    warnings.warn(f"Failed to get panorama IDs after 5 attempts: {e}")
                    return []
                print(f"Attempt {attempt + 1} failed: {e}")
                continue

    def _filter_pids_date(self, pid_df, start_date, end_date):
        """Filter panorama IDs by date."""
        pid_df["date"] = pd.to_datetime(pid_df["timestamp"], unit="ms")
        return pid_df[(pid_df["date"] >= start_date) & (pid_df["date"] <= end_date)]

    def _save_image(self, pid, data, cropped):
        """Save an image to disk."""
        img_path = os.path.join(self.dir_output, f"{pid}.jpg")
        try:
            proxy = random.choice(self._proxies)
            headers = {"User-Agent": random.choice(self._user_agents)["user_agent"]}
            image = Image.open(
                BytesIO(
                    requests.get(
                        data["_links"]["equirectangular_medium"]["href"],
                        proxies=proxy,
                        headers=headers,
                    ).content
                )
            )
            if cropped:
                image = image.crop((0, 0, image.width, image.height // 2))
            image.save(img_path)
        except Exception:
            self.logger.log_failed_pids(pid)

    def download_svi(
        self,
        dir_output: str,
        path_pid: str = None,
        cropped: bool = False,
        lat: float = None,
        lon: float = None,
        input_csv_file: str = "",
        input_shp_file: str = "",
        input_place_name: str = "",
        buffer: int = 0,
        distance: int = 10,
        start_date: str = None,
        end_date: str = None,
        metadata_only: bool = False,
        grid: bool = False,
        grid_size: int = 100,
        max_workers: int = None,
    ):
        """Download street view images from Amsterdam Street View API using specified
        parameters.

        Args:
            dir_output (str): The output directory.
            path_pid (str, optional): The path to the panorama ID file.
                Defaults to None.
            cropped (bool, optional): Whether to crop the images.
                Defaults to False.
            lat (float, optional): The latitude for the images. Defaults
                to None.
            lon (float, optional): The longitude for the images.
                Defaults to None.
            input_csv_file (str, optional): The input CSV file. Defaults
                to "".
            input_shp_file (str, optional): The input shapefile.
                Defaults to "".
            input_place_name (str, optional): The input place name.
                Defaults to "".
            buffer (int, optional): The buffer size. Defaults to 0.
            distance (int, optional): The sampling distance for lines.
                Defaults to 10.
            start_date (str, optional): The start date for the panorama
                IDs. Format is isoformat (YYYY-MM-DD). Defaults to None.
            end_date (str, optional): The end date for the panorama IDs.
                Format is isoformat (YYYY- MM-DD). Defaults to None.
            metadata_only (bool, optional): Whether to download metadata
                only. Defaults to False.
            grid (bool, optional): Grid parameter for the GeoProcessor.
                Defaults to False.
            grid_size (int, optional): Grid size parameter for the
                GeoProcessor. Defaults to 1.
            max_workers (int, optional): Number of workers for parallel processing.
                If not specified, uses the value set during initialization.

        Returns:
            None

        Raises:
            ValueError: If neither lat and lon, csv file, shapefile, nor
                place name is provided.
        """
        if self.logger is not None:
            self.logger.log_args(
                "AMSDownloader download_svi",
                dir_output=dir_output,
                path_pid=path_pid,
                cropped=cropped,
                lat=lat,
                lon=lon,
                input_csv_file=input_csv_file,
                input_shp_file=input_shp_file,
                input_place_name=input_place_name,
                buffer=buffer,
                distance=distance,
                start_date=start_date,
                end_date=end_date,
                metadata_only=metadata_only,
                grid=grid,
                grid_size=grid_size,
                max_workers=max_workers,
            )

        # Set max_workers if provided
        if max_workers is not None:
            self.max_workers = max_workers

        self.grid = grid
        self.grid_size = grid_size
        self._set_dirs(dir_output)
        self.buffer = buffer
        self.distance = distance

        if input_csv_file:
            df = pd.read_csv(input_csv_file)
            df = standardize_column_names(df)
            pid_df = self._get_pids_from_df(df)
        elif input_shp_file:
            gdf = gpd.read_file(input_shp_file)
            pid_df = self._get_pids_from_df(gdf)
        elif input_place_name:
            gdf = ox.geocode_to_gdf(input_place_name)
            pid_df = self._get_pids_from_df(gdf)
        elif lat is not None and lon is not None:
            df = pd.DataFrame({"latitude": [lat], "longitude": [lon]})
            pid_df = self._get_pids_from_df(df)
        else:
            raise ValueError("Please provide either lat and lon, input_csv_file, input_shp_file, or input_place_name.")

        if start_date and end_date:
            pid_df = self._filter_pids_date(pid_df, start_date, end_date)

        def process_pid(pid, max_retries=5):
            img_url = f"https://api.data.amsterdam.nl/panorama/panoramas/{pid}/"

            for attempt in range(max_retries):
                try:
                    proxy = random.choice(self._proxies)
                    headers = {"User-Agent": random.choice(self._user_agents)["user_agent"]}
                    response = requests.get(img_url, proxies=proxy, headers=headers)
                    data = json.loads(response.content)

                    if response.status_code == 200:
                        if not metadata_only:
                            self._save_image(pid, data, cropped)

                        return {
                            "geometry": Point(data["geometry"]["coordinates"][:2]),
                            "lat": data["geometry"]["coordinates"][1],
                            "lon": data["geometry"]["coordinates"][0],
                            "pano_id": data["pano_id"],
                            "timestamp": data["timestamp"],
                            "mission_year": data["mission_year"],
                            "roll": data["roll"],
                            "pitch": data["pitch"],
                            "heading": data["heading"],
                        }
                except Exception as e:
                    if attempt == max_retries - 1:  # Last attempt
                        print(f"Failed to download data for PID {pid} after {max_retries} attempts: {str(e)}")
                        return None
                    continue
            return None

        results = []
        with ThreadPoolExecutor(max_workers=self.max_workers) as executor:
            futures = [executor.submit(process_pid, row.pano_id) for _, row in pid_df.iterrows()]
            for future in tqdm(
                as_completed(futures),
                total=len(futures),
                desc="Downloading images and metadata",
            ):
                result = future.result()
                if result is not None:
                    results.append(result)
        if len(results) == 0:
            print("No data downloaded. Please check the input parameters.")
            return

        final_df = gpd.GeoDataFrame(results, crs="EPSG:4326", geometry="geometry")
        final_df.to_csv(os.path.join(self.dir_output, "ams_pids.csv"), index=False)
        print(f"Metadata saved to {os.path.join(self.dir_output, 'ams_pids.csv')}")



"""Transforms module for depth estimation.

This module contains the transforms used for preprocessing images for depth estimation.
"""

import numpy as np
import torch.nn as nn
from torchvision.transforms import Normalize


class Resize(object):
    """Resize sample to given size (width, height)."""

    def __init__(
        self,
        width,
        height,
        resize_target=True,
        keep_aspect_ratio=False,
        ensure_multiple_of=1,
        resize_method="lower_bound",
    ):
        """Initializes the Resize object.

        Args:
            width (int): Desired output width.
            height (int): Desired output height.
            resize_target (bool, optional): If True, resize the full sample (image, mask, target). Defaults to True.
            keep_aspect_ratio (bool, optional): If True, keep the aspect ratio of the input sample. Defaults to False.
            ensure_multiple_of (int, optional): Output width and height is constrained to be multiple of this parameter. Defaults to 1.
            resize_method (str, optional): Method of resizing. Options are "lower_bound", "upper_bound", "minimal". Defaults to "lower_bound".
        """
        self.__width = width
        self.__height = height
        self.__keep_aspect_ratio = keep_aspect_ratio
        self.__multiple_of = ensure_multiple_of
        self.__resize_method = resize_method

    def constrain_to_multiple_of(self, x, min_val=0, max_val=None):
        """Constrain a value to be a multiple of a specified number.

        Args:
            x (float): Value to constrain.
            min_val (int, optional): Minimum value. Defaults to 0.
            max_val (int, optional): Maximum value. Defaults to None.

        Returns:
            int: Constrained value.
        """
        y = (np.round(x / self.__multiple_of) * self.__multiple_of).astype(int)

        if max_val is not None and y > max_val:
            y = (np.floor(x / self.__multiple_of) * self.__multiple_of).astype(int)

        if y < min_val:
            y = (np.ceil(x / self.__multiple_of) * self.__multiple_of).astype(int)

        return y

    def get_size(self, width, height):
        """Determine the new size based on the resizing method.

        Args:
            width (int): Original width.
            height (int): Original height.

        Returns:
            tuple: New width and height.
        """
        # determine new height and width
        scale_height = self.__height / height
        scale_width = self.__width / width

        if self.__keep_aspect_ratio:
            if self.__resize_method == "lower_bound":
                # scale such that output size is lower bound
                if scale_width > scale_height:
                    # fit width
                    scale_height = scale_width
                else:
                    # fit height
                    scale_width = scale_height
            elif self.__resize_method == "upper_bound":
                # scale such that output size is upper bound
                if scale_width < scale_height:
                    # fit width
                    scale_height = scale_width
                else:
                    # fit height
                    scale_width = scale_height
            elif self.__resize_method == "minimal":
                # scale as least as possible
                if abs(1 - scale_width) < abs(1 - scale_height):
                    # fit width
                    scale_height = scale_width
                else:
                    # fit height
                    scale_width = scale_height
            else:
                raise ValueError(f"resize_method {self.__resize_method} not implemented")

        if self.__resize_method == "lower_bound":
            new_height = self.constrain_to_multiple_of(scale_height * height, min_val=self.__height)
            new_width = self.constrain_to_multiple_of(scale_width * width, min_val=self.__width)
        elif self.__resize_method == "upper_bound":
            new_height = self.constrain_to_multiple_of(scale_height * height, max_val=self.__height)
            new_width = self.constrain_to_multiple_of(scale_width * width, max_val=self.__width)
        elif self.__resize_method == "minimal":
            new_height = self.constrain_to_multiple_of(scale_height * height)
            new_width = self.constrain_to_multiple_of(scale_width * width)
        else:
            raise ValueError(f"resize_method {self.__resize_method} not implemented")

        return (new_width, new_height)

    def __call__(self, x):
        """Resize the input tensor.

        Args:
            x (torch.Tensor): Input tensor of shape (N, C, H, W).

        Returns:
            torch.Tensor: Resized tensor.
        """
        width, height = self.get_size(*x.shape[-2:][::-1])
        return nn.functional.interpolate(x, (height, width), mode="bilinear", align_corners=True)


class PrepareForNet(object):
    """Prepares input for network by resizing and normalizing."""

    def __init__(
        self,
        resize_mode="minimal",
        keep_aspect_ratio=True,
        img_size=384,
        do_resize=True,
    ):
        """Initializes the PrepareForNet object.

        Args:
            resize_mode (str, optional): Method of resizing. Defaults to "minimal".
            keep_aspect_ratio (bool, optional): If True, keep the aspect ratio of the input image. Defaults to True.
            img_size (int, tuple, optional): Size of the input image. Defaults to 384.
            do_resize (bool, optional): If True, perform resizing. Defaults to True.
        """
        if isinstance(img_size, int):
            img_size = (img_size, img_size)
        net_h, net_w = img_size
        self.normalization = Normalize(mean=[0.485, 0.456, 0.406], std=[0.229, 0.224, 0.225])
        self.resizer = (
            Resize(
                net_w,
                net_h,
                keep_aspect_ratio=keep_aspect_ratio,
                ensure_multiple_of=14,
                resize_method=resize_mode,
            )
            if do_resize
            else nn.Identity()
        )

    def __call__(self, x):
        """Apply normalization and resizing to the input tensor.

        Args:
            x (torch.Tensor): Input tensor of shape (N, C, H, W).

        Returns:
            torch.Tensor: Normalized and resized tensor.
        """
        return self.normalization(self.resizer(x))


# Alias for backward compatibility
NormalizeImage = Normalize



import argparse

import torch
import torch.nn as nn
import torch.nn.functional as F
from depth_anything.blocks import FeatureFusionBlock, _make_scratch
from huggingface_hub import PyTorchModelHubMixin


def _make_fusion_block(features, use_bn, size=None):
    """Creates a feature fusion block.

    Args:
        features (int): Number of input features.
        use_bn (bool): Whether to use batch normalization.
        size (tuple, optional): Size of the output. Defaults to None.

    Returns:
        FeatureFusionBlock: The constructed feature fusion block.
    """
    return FeatureFusionBlock(
        features,
        nn.ReLU(False),
        deconv=False,
        bn=use_bn,
        expand=False,
        align_corners=True,
        size=size,
    )


class DPTHead(nn.Module):
    """DPT Head for depth estimation."""

    def __init__(
        self,
        nclass,
        in_channels,
        features=256,
        use_bn=False,
        out_channels=[256, 512, 1024, 1024],
        use_clstoken=False,
    ):
        """Initializes the DPTHead.

        Args:
            nclass (int): Number of output classes.
            in_channels (int): Number of input channels.
            features (int, optional): Number of features. Defaults to 256.
            use_bn (bool, optional): Whether to use batch normalization. Defaults to False.
            out_channels (list, optional): List of output channels. Defaults to [256, 512, 1024, 1024].
            use_clstoken (bool, optional): Whether to use class token. Defaults to False.
        """
        super(DPTHead, self).__init__()

        self.nclass = nclass
        self.use_clstoken = use_clstoken

        self.projects = nn.ModuleList(
            [
                nn.Conv2d(
                    in_channels=in_channels,
                    out_channels=out_channel,
                    kernel_size=1,
                    stride=1,
                    padding=0,
                )
                for out_channel in out_channels
            ]
        )

        self.resize_layers = nn.ModuleList(
            [
                nn.ConvTranspose2d(
                    in_channels=out_channels[0],
                    out_channels=out_channels[0],
                    kernel_size=4,
                    stride=4,
                    padding=0,
                ),
                nn.ConvTranspose2d(
                    in_channels=out_channels[1],
                    out_channels=out_channels[1],
                    kernel_size=2,
                    stride=2,
                    padding=0,
                ),
                nn.Identity(),
                nn.Conv2d(
                    in_channels=out_channels[3],
                    out_channels=out_channels[3],
                    kernel_size=3,
                    stride=2,
                    padding=1,
                ),
            ]
        )

        if use_clstoken:
            self.readout_projects = nn.ModuleList()
            for _ in range(len(self.projects)):
                self.readout_projects.append(nn.Sequential(nn.Linear(2 * in_channels, in_channels), nn.GELU()))

        self.scratch = _make_scratch(
            out_channels,
            features,
            groups=1,
            expand=False,
        )

        self.scratch.stem_transpose = None

        self.scratch.refinenet1 = _make_fusion_block(features, use_bn)
        self.scratch.refinenet2 = _make_fusion_block(features, use_bn)
        self.scratch.refinenet3 = _make_fusion_block(features, use_bn)
        self.scratch.refinenet4 = _make_fusion_block(features, use_bn)

        head_features_1 = features
        head_features_2 = 32

        if nclass > 1:
            self.scratch.output_conv = nn.Sequential(
                nn.Conv2d(head_features_1, head_features_1, kernel_size=3, stride=1, padding=1),
                nn.ReLU(True),
                nn.Conv2d(head_features_1, nclass, kernel_size=1, stride=1, padding=0),
            )
        else:
            self.scratch.output_conv1 = nn.Conv2d(
                head_features_1,
                head_features_1 // 2,
                kernel_size=3,
                stride=1,
                padding=1,
            )

            self.scratch.output_conv2 = nn.Sequential(
                nn.Conv2d(
                    head_features_1 // 2,
                    head_features_2,
                    kernel_size=3,
                    stride=1,
                    padding=1,
                ),
                nn.ReLU(True),
                nn.Conv2d(head_features_2, 1, kernel_size=1, stride=1, padding=0),
                nn.ReLU(True),
                nn.Identity(),
            )

    def forward(self, out_features, patch_h, patch_w):
        """Forward pass for the DPTHead.

        Args:
            out_features (list): List of output features from the encoder.
            patch_h (int): Height of the patch.
            patch_w (int): Width of the patch.

        Returns:
            Tensor: The output tensor after processing through the head.
        """
        out = []
        for i, x in enumerate(out_features):
            if self.use_clstoken:
                x, cls_token = x[0], x[1]
                readout = cls_token.unsqueeze(1).expand_as(x)
                x = self.readout_projects[i](torch.cat((x, readout), -1))
            else:
                x = x[0]

            x = x.permute(0, 2, 1).reshape((x.shape[0], x.shape[-1], patch_h, patch_w))

            x = self.projects[i](x)
            x = self.resize_layers[i](x)

            out.append(x)

        layer_1, layer_2, layer_3, layer_4 = out

        layer_1_rn = self.scratch.layer1_rn(layer_1)
        layer_2_rn = self.scratch.layer2_rn(layer_2)
        layer_3_rn = self.scratch.layer3_rn(layer_3)
        layer_4_rn = self.scratch.layer4_rn(layer_4)

        path_4 = self.scratch.refinenet4(layer_4_rn, size=layer_3_rn.shape[2:])
        path_3 = self.scratch.refinenet3(path_4, layer_3_rn, size=layer_2_rn.shape[2:])
        path_2 = self.scratch.refinenet2(path_3, layer_2_rn, size=layer_1_rn.shape[2:])
        path_1 = self.scratch.refinenet1(path_2, layer_1_rn)

        out = self.scratch.output_conv1(path_1)
        out = F.interpolate(
            out,
            (int(patch_h * 14), int(patch_w * 14)),
            mode="bilinear",
            align_corners=True,
        )
        out = self.scratch.output_conv2(out)

        return out


class DPT_DINOv2(nn.Module):
    """DPT model using DINOv2 for depth estimation."""

    def __init__(
        self,
        encoder="vitl",
        features=256,
        out_channels=[256, 512, 1024, 1024],
        use_bn=False,
        use_clstoken=False,
        localhub=True,
    ):
        """Initializes the DPT_DINOv2 model.

        Args:
            encoder (str, optional): The encoder type. Defaults to "vitl".
            features (int, optional): Number of features. Defaults to 256.
            out_channels (list, optional): List of output channels. Defaults to [256, 512, 1024, 1024].
            use_bn (bool, optional): Whether to use batch normalization. Defaults to False.
            use_clstoken (bool, optional): Whether to use class token. Defaults to False.
            localhub (bool, optional): Whether to load the model locally. Defaults to True.
        """
        super(DPT_DINOv2, self).__init__()

        assert encoder in ["vits", "vitb", "vitl"]

        # in case the Internet connection is not stable, please load the DINOv2 locally
        if localhub:
            self.pretrained = torch.hub.load(
                "torchhub/facebookresearch_dinov2_main",
                "dinov2_{:}14".format(encoder),
                source="local",
                pretrained=False,
            )
        else:
            self.pretrained = torch.hub.load("facebookresearch/dinov2", "dinov2_{:}14".format(encoder))

        dim = self.pretrained.blocks[0].attn.qkv.in_features

        self.depth_head = DPTHead(
            1,
            dim,
            features,
            use_bn,
            out_channels=out_channels,
            use_clstoken=use_clstoken,
        )

    def forward(self, x):
        """Forward pass for the DPT_DINOv2 model.

        Args:
            x (Tensor): Input tensor.

        Returns:
            Tensor: The estimated depth map.
        """
        h, w = x.shape[-2:]

        features = self.pretrained.get_intermediate_layers(x, 4, return_class_token=True)

        patch_h, patch_w = h // 14, w // 14

        depth = self.depth_head(features, patch_h, patch_w)
        depth = F.interpolate(depth, size=(h, w), mode="bilinear", align_corners=True)
        depth = F.relu(depth)

        return depth.squeeze(1)


class DepthAnything(DPT_DINOv2, PyTorchModelHubMixin):
    """DepthAnything model for depth estimation."""

    def __init__(self, config):
        """Initializes the DepthAnything model.

        Args:
            config (dict): Configuration dictionary for the model.
        """
        super().__init__(**config)


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--encoder",
        default="vits",
        type=str,
        choices=["vits", "vitb", "vitl"],
    )
    args = parser.parse_args()

    model = DepthAnything.from_pretrained("LiheYoung/depth_anything_{:}14".format(args.encoder))

    print(model)



from torch import nn
from torchvision.models import ViT_B_16_Weights, vit_b_16


class Net(nn.Module):
    """Neural network model for human perception prediction using Vision Transformer.

    This model was developed by Ouyan (2023) and uses a pre-trained ViT-B-16 backbone
    with custom classification head for predicting human perception scores of places.
    The original code can be accessed at:
    https://github.com/strawmelon11/human-perception-place-pulse

    Args:
        num_classes: Number of output classes for classification.

    Attributes:
        model: The Vision Transformer model with custom classification head.
    """

    def __init__(self, num_classes):
        super(Net, self).__init__()

        self.model = vit_b_16(weights=ViT_B_16_Weights.IMAGENET1K_SWAG_E2E_V1)
        num_fc = self.model.heads.head.in_features
        self.model.heads.head = nn.Sequential(
            nn.Linear(num_fc, 512, bias=True),
            nn.ReLU(True),
            nn.Linear(512, 256, bias=True),
            nn.ReLU(True),
            nn.Linear(256, num_classes, bias=True),
        )
        nn.init.xavier_uniform_(self.model.heads.head[0].weight)
        nn.init.xavier_uniform_(self.model.heads.head[2].weight)
        nn.init.xavier_uniform_(self.model.heads.head[4].weight)

    def forward(self, x):
        """Forward pass of the model.

        Args:
            x: Input tensor of shape (batch_size, channels, height, width).

        Returns:
            Tensor of perception scores scaled between 0-10.
        """
        x = self.model(x)

        scores = self.calculate_score(x)
        return scores

    def calculate_score(self, probabilities):
        """Calculates final perception scores from model outputs.

        Args:
            probabilities: Raw model output logits.

        Returns:
            Tensor of perception scores scaled between 0-10 and rounded to 2 decimals.
        """
        softmax = nn.Softmax(dim=1)
        scores = softmax(probabilities)[:, 1]  # get all probabilities
        scores = (scores * 10).round(decimals=2)  # scale and round values

        return scores



# Copyright (c) Meta Platforms, Inc. and affiliates.
# All rights reserved.
#
# This source code is licensed under the license found in the
# LICENSE file in the root directory of this source tree.

# References:
#   https://github.com/facebookresearch/dino/blob/master/vision_transformer.py
#   https://github.com/rwightman/pytorch-image-models/tree/master/timm/layers/mlp.py


from typing import Callable, Optional

from torch import Tensor, nn


class Mlp(nn.Module):
    """Multi-layer perceptron (MLP) module.

    This module implements a simple feedforward neural network with
    one hidden layer, an activation function, and dropout.

    Attributes:
        fc1 (nn.Linear): The first linear layer.
        act (Callable): The activation function.
        fc2 (nn.Linear): The second linear layer.
        drop (nn.Dropout): The dropout layer.

    Args:
        in_features (int): Number of input features.
        hidden_features (Optional[int]): Number of hidden features. If None, defaults to in_features.
        out_features (Optional[int]): Number of output features. If None, defaults to in_features.
        act_layer (Callable[..., nn.Module]): Activation layer to use. Defaults to nn.GELU.
        drop (float): Dropout probability. Defaults to 0.0.
        bias (bool): Whether to use bias in the linear layers. Defaults to True.
    """

    def __init__(
        self,
        in_features: int,
        hidden_features: Optional[int] = None,
        out_features: Optional[int] = None,
        act_layer: Callable[..., nn.Module] = nn.GELU,
        drop: float = 0.0,
        bias: bool = True,
    ) -> None:
        super().__init__()
        out_features = out_features or in_features
        hidden_features = hidden_features or in_features
        self.fc1 = nn.Linear(in_features, hidden_features, bias=bias)
        self.act = act_layer()
        self.fc2 = nn.Linear(hidden_features, out_features, bias=bias)
        self.drop = nn.Dropout(drop)

    def forward(self, x: Tensor) -> Tensor:
        """Forward pass through the MLP.

        Args:
            x (Tensor): Input tensor of shape (N, in_features), where N is the batch size.

        Returns:
            Tensor: Output tensor of shape (N, out_features).
        """
        x = self.fc1(x)
        x = self.act(x)
        x = self.drop(x)
        x = self.fc2(x)
        x = self.drop(x)
        return x



# Copyright (c) Meta Platforms, Inc. and affiliates.
# All rights reserved.
#
# This source code is licensed under the license found in the
# LICENSE file in the root directory of this source tree.

import csv
import logging
import os
from enum import Enum
from typing import Callable, List, Optional, Tuple, Union

import numpy as np

from .extended import ExtendedVisionDataset

logger = logging.getLogger("dinov2")
_Target = int


class _Split(Enum):
    """Enumeration for dataset splits."""

    TRAIN = "train"
    VAL = "val"
    TEST = "test"  # NOTE: torchvision does not support the test split

    @property
    def length(self) -> int:
        """Returns the number of samples in the split.

        Returns:
            int: The number of samples in the split.
        """
        split_lengths = {
            _Split.TRAIN: 1_281_167,
            _Split.VAL: 50_000,
            _Split.TEST: 100_000,
        }
        return split_lengths[self]

    def get_dirname(self, class_id: Optional[str] = None) -> str:
        """Gets the directory name for the split.

        Args:
            class_id (Optional[str]): The class ID to include in the path. Defaults to None.

        Returns:
            str: The directory name for the split.
        """
        return self.value if class_id is None else os.path.join(self.value, class_id)

    def get_image_relpath(self, actual_index: int, class_id: Optional[str] = None) -> str:
        """Generates the relative path for an image.

        Args:
            actual_index (int): The actual index of the image.
            class_id (Optional[str]): The class ID. Defaults to None.

        Returns:
            str: The relative path to the image.
        """
        dirname = self.get_dirname(class_id)
        if self == _Split.TRAIN:
            basename = f"{class_id}_{actual_index}"
        else:  # self in (_Split.VAL, _Split.TEST):
            basename = f"ILSVRC2012_{self.value}_{actual_index:08d}"
        return os.path.join(dirname, basename + ".JPEG")

    def parse_image_relpath(self, image_relpath: str) -> Tuple[str, int]:
        """Parses the image relative path to extract class ID and actual index.

        Args:
            image_relpath (str): The relative path of the image.

        Returns:
            Tuple[str, int]: A tuple containing the class ID and actual index.
        """
        assert self != _Split.TEST
        dirname, filename = os.path.split(image_relpath)
        class_id = os.path.split(dirname)[-1]
        basename, _ = os.path.splitext(filename)
        actual_index = int(basename.split("_")[-1])
        return class_id, actual_index


class ImageNet(ExtendedVisionDataset):
    """ImageNet dataset class extending the ExtendedVisionDataset.

    Attributes:
        Target (Union[_Target]): Type for target values.
        Split (Union[_Split]): Type for dataset splits.
    """

    Target = Union[_Target]
    Split = Union[_Split]

    def __init__(
        self,
        *,
        split: "ImageNet.Split",
        root: str,
        extra: str,
        transforms: Optional[Callable] = None,
        transform: Optional[Callable] = None,
        target_transform: Optional[Callable] = None,
    ) -> None:
        """Initializes the ImageNet dataset.

        Args:
            split (ImageNet.Split): The dataset split (train, val, test).
            root (str): The root directory of the dataset.
            extra (str): The directory for extra data.
            transforms (Optional[Callable]): Transformations to apply to the images. Defaults to None.
            transform (Optional[Callable]): Transformations to apply to the images. Defaults to None.
            target_transform (Optional[Callable]): Transformations to apply to the targets. Defaults to None.
        """
        super().__init__(root, transforms, transform, target_transform)
        self._extra_root = extra
        self._split = split

        self._entries = None
        self._class_ids = None
        self._class_names = None

    @property
    def split(self) -> "ImageNet.Split":
        """Returns the dataset split.

        Returns:
            ImageNet.Split: The dataset split.
        """
        return self._split

    def _get_extra_full_path(self, extra_path: str) -> str:
        """Constructs the full path for extra data.

        Args:
            extra_path (str): The extra path to be appended.

        Returns:
            str: The full path to the extra data.
        """
        return os.path.join(self._extra_root, extra_path)

    def _load_extra(self, extra_path: str) -> np.ndarray:
        """Loads extra data from the specified path.

        Args:
            extra_path (str): The path to the extra data.

        Returns:
            np.ndarray: The loaded extra data.
        """
        extra_full_path = self._get_extra_full_path(extra_path)
        return np.load(extra_full_path, mmap_mode="r")

    def _save_extra(self, extra_array: np.ndarray, extra_path: str) -> None:
        """Saves extra data to the specified path.

        Args:
            extra_array (np.ndarray): The extra data to save.
            extra_path (str): The path where the data will be saved.
        """
        extra_full_path = self._get_extra_full_path(extra_path)
        os.makedirs(self._extra_root, exist_ok=True)
        np.save(extra_full_path, extra_array)

    @property
    def _entries_path(self) -> str:
        """Returns the path for the entries file.

        Returns:
            str: The path for the entries file.
        """
        return f"entries-{self._split.value.upper()}.npy"

    @property
    def _class_ids_path(self) -> str:
        """Returns the path for the class IDs file.

        Returns:
            str: The path for the class IDs file.
        """
        return f"class-ids-{self._split.value.upper()}.npy"

    @property
    def _class_names_path(self) -> str:
        """Returns the path for the class names file.

        Returns:
            str: The path for the class names file.
        """
        return f"class-names-{self._split.value.upper()}.npy"

    def _get_entries(self) -> np.ndarray:
        """Retrieves the entries for the dataset.

        Returns:
            np.ndarray: The entries for the dataset.
        """
        if self._entries is None:
            self._entries = self._load_extra(self._entries_path)
        assert self._entries is not None
        return self._entries

    def _get_class_ids(self) -> np.ndarray:
        """Retrieves the class IDs for the dataset.

        Returns:
            np.ndarray: The class IDs for the dataset.

        Raises:
            AssertionError: If the split is TEST, as class IDs are not available.
        """
        if self._split == _Split.TEST:
            assert False, "Class IDs are not available in TEST split"
        if self._class_ids is None:
            self._class_ids = self._load_extra(self._class_ids_path)
        assert self._class_ids is not None
        return self._class_ids

    def _get_class_names(self) -> np.ndarray:
        """Retrieves the class names for the dataset.

        Returns:
            np.ndarray: The class names for the dataset.

        Raises:
            AssertionError: If the split is TEST, as class names are not available.
        """
        if self._split == _Split.TEST:
            assert False, "Class names are not available in TEST split"
        if self._class_names is None:
            self._class_names = self._load_extra(self._class_names_path)
        assert self._class_names is not None
        return self._class_names

    def find_class_id(self, class_index: int) -> str:
        """Finds the class ID for a given class index.

        Args:
            class_index (int): The index of the class.

        Returns:
            str: The class ID corresponding to the class index.
        """
        class_ids = self._get_class_ids()
        return str(class_ids[class_index])

    def find_class_name(self, class_index: int) -> str:
        """Finds the class name for a given class index.

        Args:
            class_index (int): The index of the class.

        Returns:
            str: The class name corresponding to the class index.
        """
        class_names = self._get_class_names()
        return str(class_names[class_index])

    def get_image_data(self, index: int) -> bytes:
        """Retrieves the image data for a given index.

        Args:
            index (int): The index of the image.

        Returns:
            bytes: The image data as bytes.
        """
        entries = self._get_entries()
        actual_index = entries[index]["actual_index"]

        class_id = self.get_class_id(index)

        image_relpath = self.split.get_image_relpath(actual_index, class_id)
        image_full_path = os.path.join(self.root, image_relpath)
        with open(image_full_path, mode="rb") as f:
            image_data = f.read()
        return image_data

    def get_target(self, index: int) -> Optional[Target]:
        """Retrieves the target for a given index.

        Args:
            index (int): The index of the target.

        Returns:
            Optional[Target]: The target value, or None if the split is TEST.
        """
        entries = self._get_entries()
        class_index = entries[index]["class_index"]
        return None if self.split == _Split.TEST else int(class_index)

    def get_targets(self) -> Optional[np.ndarray]:
        """Retrieves all targets for the dataset.

        Returns:
            Optional[np.ndarray]: The targets for the dataset, or None if the split is TEST.
        """
        entries = self._get_entries()
        return None if self.split == _Split.TEST else entries["class_index"]

    def get_class_id(self, index: int) -> Optional[str]:
        """Retrieves the class ID for a given index.

        Args:
            index (int): The index of the class.

        Returns:
            Optional[str]: The class ID, or None if the split is TEST.
        """
        entries = self._get_entries()
        class_id = entries[index]["class_id"]
        return None if self.split == _Split.TEST else str(class_id)

    def get_class_name(self, index: int) -> Optional[str]:
        """Retrieves the class name for a given index.

        Args:
            index (int): The index of the class.

        Returns:
            Optional[str]: The class name, or None if the split is TEST.
        """
        entries = self._get_entries()
        class_name = entries[index]["class_name"]
        return None if self.split == _Split.TEST else str(class_name)

    def __len__(self) -> int:
        """Returns the number of entries in the dataset.

        Returns:
            int: The number of entries in the dataset.
        """
        entries = self._get_entries()
        assert len(entries) == self.split.length
        return len(entries)

    def _load_labels(self, labels_path: str) -> List[Tuple[str, str]]:
        """Loads labels from a specified file.

        Args:
            labels_path (str): The path to the labels file.

        Returns:
            List[Tuple[str, str]]: A list of tuples containing class IDs and class names.

        Raises:
            RuntimeError: If the labels file cannot be read.
        """
        labels_full_path = os.path.join(self.root, labels_path)
        labels = []

        try:
            with open(labels_full_path, "r") as f:
                reader = csv.reader(f)
                for row in reader:
                    class_id, class_name = row
                    labels.append((class_id, class_name))
        except OSError as e:
            raise RuntimeError(f'can not read labels file "{labels_full_path}"') from e

        return labels

    def _dump_entries(self) -> None:
        """Dumps the entries to a file based on the dataset split."""
        split = self.split
        if split == ImageNet.Split.TEST:
            dataset = None
            sample_count = split.length
            max_class_id_length, max_class_name_length = 0, 0
        else:
            labels_path = "labels.txt"
            logger.info(f'loading labels from "{labels_path}"')
            labels = self._load_labels(labels_path)

            # NOTE: Using torchvision ImageFolder for consistency
            from torchvision.datasets import ImageFolder

            dataset_root = os.path.join(self.root, split.get_dirname())
            dataset = ImageFolder(dataset_root)
            sample_count = len(dataset)
            max_class_id_length, max_class_name_length = -1, -1
            for sample in dataset.samples:
                _, class_index = sample
                class_id, class_name = labels[class_index]
                max_class_id_length = max(len(class_id), max_class_id_length)
                max_class_name_length = max(len(class_name), max_class_name_length)

        dtype = np.dtype(
            [
                ("actual_index", "<u4"),
                ("class_index", "<u4"),
                ("class_id", f"U{max_class_id_length}"),
                ("class_name", f"U{max_class_name_length}"),
            ]
        )
        entries_array = np.empty(sample_count, dtype=dtype)

        if split == ImageNet.Split.TEST:
            old_percent = -1
            for index in range(sample_count):
                percent = 100 * (index + 1) // sample_count
                if percent > old_percent:
                    logger.info(f"creating entries: {percent}%")
                    old_percent = percent

                actual_index = index + 1
                class_index = np.uint32(-1)
                class_id, class_name = "", ""
                entries_array[index] = (actual_index, class_index, class_id, class_name)
        else:
            class_names = {class_id: class_name for class_id, class_name in labels}

            assert dataset
            old_percent = -1
            for index in range(sample_count):
                percent = 100 * (index + 1) // sample_count
                if percent > old_percent:
                    logger.info(f"creating entries: {percent}%")
                    old_percent = percent

                image_full_path, class_index = dataset.samples[index]
                image_relpath = os.path.relpath(image_full_path, self.root)
                class_id, actual_index = split.parse_image_relpath(image_relpath)
                class_name = class_names[class_id]
                entries_array[index] = (actual_index, class_index, class_id, class_name)

        logger.info(f'saving entries to "{self._entries_path}"')
        self._save_extra(entries_array, self._entries_path)

    def _dump_class_ids_and_names(self) -> None:
        """Dumps class IDs and names to files based on the dataset split."""
        split = self.split
        if split == ImageNet.Split.TEST:
            return

        entries_array = self._load_extra(self._entries_path)

        max_class_id_length, max_class_name_length, max_class_index = -1, -1, -1
        for entry in entries_array:
            class_index, class_id, class_name = (
                entry["class_index"],
                entry["class_id"],
                entry["class_name"],
            )
            max_class_index = max(int(class_index), max_class_index)
            max_class_id_length = max(len(str(class_id)), max_class_id_length)
            max_class_name_length = max(len(str(class_name)), max_class_name_length)

        class_count = max_class_index + 1
        class_ids_array = np.empty(class_count, dtype=f"U{max_class_id_length}")
        class_names_array = np.empty(class_count, dtype=f"U{max_class_name_length}")
        for entry in entries_array:
            class_index, class_id, class_name = (
                entry["class_index"],
                entry["class_id"],
                entry["class_name"],
            )
            class_ids_array[class_index] = class_id
            class_names_array[class_index] = class_name

        logger.info(f'saving class IDs to "{self._class_ids_path}"')
        self._save_extra(class_ids_array, self._class_ids_path)

        logger.info(f'saving class names to "{self._class_names_path}"')
        self._save_extra(class_names_array, self._class_names_path)

    def dump_extra(self) -> None:
        """Dumps extra data (entries, class IDs, and class names) to files."""
        self._dump_entries()
        self._dump_class_ids_and_names()



# Copyright (c) Meta Platforms, Inc. and affiliates.
# All rights reserved.
#
# This source code is licensed under the license found in the
# LICENSE file in the root directory of this source tree.

# Modified from: https://github.com/huggingface/pytorch-image-models/blob/main/timm/models/vision_transformer.py#L103-L110

from typing import Union

import torch
from torch import Tensor, nn


class LayerScale(nn.Module):
    """Applies layer scaling to the input tensor.

    This module scales the input tensor by a learnable parameter (gamma),
    which can be initialized to a specific value. The scaling can be done
    in-place or not, depending on the `inplace` parameter.

    Attributes:
        gamma (nn.Parameter): The learnable scaling parameter.

    Args:
        dim (int): The dimension of the input tensor.
        init_values (Union[float, Tensor], optional): Initial value(s) for gamma. Defaults to 1e-5.
        inplace (bool, optional): If True, performs the operation in-place. Defaults to False.
    """

    def __init__(
        self,
        dim: int,
        init_values: Union[float, Tensor] = 1e-5,
        inplace: bool = False,
    ) -> None:
        super().__init__()
        self.inplace = inplace
        self.gamma = nn.Parameter(init_values * torch.ones(dim))

    def forward(self, x: Tensor) -> Tensor:
        """Forward pass for the LayerScale module.

        Args:
            x (Tensor): The input tensor to be scaled.

        Returns:
            Tensor: The scaled tensor.
        """
        return x.mul_(self.gamma) if self.inplace else x * self.gamma



# Copyright (c) Meta Platforms, Inc. and affiliates.
# All rights reserved.
#
# This source code is licensed under the license found in the
# LICENSE file in the root directory of this source tree.

import logging
from enum import Enum
from typing import Any, Dict, Optional

import torch
from torch import Tensor
from torchmetrics import Metric, MetricCollection
from torchmetrics.classification import MulticlassAccuracy
from torchmetrics.utilities.data import dim_zero_cat, select_topk

logger = logging.getLogger("dinov2")


class MetricType(Enum):
    """Enumeration of different metric types for evaluation."""

    MEAN_ACCURACY = "mean_accuracy"
    MEAN_PER_CLASS_ACCURACY = "mean_per_class_accuracy"
    PER_CLASS_ACCURACY = "per_class_accuracy"
    IMAGENET_REAL_ACCURACY = "imagenet_real_accuracy"

    @property
    def accuracy_averaging(self):
        """Returns the accuracy averaging method associated with the metric type."""
        return getattr(AccuracyAveraging, self.name, None)

    def __str__(self):
        return self.value


class AccuracyAveraging(Enum):
    """Enumeration of accuracy averaging methods."""

    MEAN_ACCURACY = "micro"
    MEAN_PER_CLASS_ACCURACY = "macro"
    PER_CLASS_ACCURACY = "none"

    def __str__(self):
        return self.value


def build_metric(metric_type: MetricType, *, num_classes: int, ks: Optional[tuple] = None):
    """Builds a metric based on the specified metric type.

    Args:
        metric_type (MetricType): The type of metric to build.
        num_classes (int): The number of classes for the metric.
        ks (Optional[tuple], optional): The top-k values to consider. Defaults to (1, 5).

    Returns:
        MetricCollection: A collection of metrics based on the specified type.

    Raises:
        ValueError: If the metric type is unknown.
    """
    if metric_type.accuracy_averaging is not None:
        return build_topk_accuracy_metric(
            average_type=metric_type.accuracy_averaging,
            num_classes=num_classes,
            ks=(1, 5) if ks is None else ks,
        )
    elif metric_type == MetricType.IMAGENET_REAL_ACCURACY:
        return build_topk_imagenet_real_accuracy_metric(
            num_classes=num_classes,
            ks=(1, 5) if ks is None else ks,
        )

    raise ValueError(f"Unknown metric type {metric_type}")


def build_topk_accuracy_metric(average_type: AccuracyAveraging, num_classes: int, ks: tuple = (1, 5)):
    """Builds a top-k accuracy metric.

    Args:
        average_type (AccuracyAveraging): The type of averaging to use.
        num_classes (int): The number of classes for the metric.
        ks (tuple, optional): The top-k values to consider. Defaults to (1, 5).

    Returns:
        MetricCollection: A collection of top-k accuracy metrics.
    """
    metrics: Dict[str, Metric] = {
        f"top-{k}": MulticlassAccuracy(top_k=k, num_classes=int(num_classes), average=average_type.value) for k in ks
    }
    return MetricCollection(metrics)


def build_topk_imagenet_real_accuracy_metric(num_classes: int, ks: tuple = (1, 5)):
    """Builds a top-k ImageNet real accuracy metric.

    Args:
        num_classes (int): The number of classes for the metric.
        ks (tuple, optional): The top-k values to consider. Defaults to (1, 5).

    Returns:
        MetricCollection: A collection of top-k ImageNet real accuracy metrics.
    """
    metrics: Dict[str, Metric] = {f"top-{k}": ImageNetReaLAccuracy(top_k=k, num_classes=int(num_classes)) for k in ks}
    return MetricCollection(metrics)


class ImageNetReaLAccuracy(Metric):
    """Metric for calculating ImageNet real accuracy."""

    is_differentiable: bool = False
    higher_is_better: Optional[bool] = None
    full_state_update: bool = False

    def __init__(
        self,
        num_classes: int,
        top_k: int = 1,
        **kwargs: Any,
    ) -> None:
        """Initializes the ImageNet real accuracy metric.

        Args:
            num_classes (int): The number of classes for the metric.
            top_k (int, optional): The number of top predictions to consider. Defaults to 1.
        """
        super().__init__(**kwargs)
        self.num_classes = num_classes
        self.top_k = top_k
        self.add_state("tp", [], dist_reduce_fx="cat")

    def update(self, preds: Tensor, target: Tensor) -> None:  # type: ignore
        """Updates the metric with predictions and targets.

        Args:
            preds (Tensor): The predicted values.
            target (Tensor): The ground truth values.
        """
        # preds [B, D]
        # target [B, A]
        # preds_oh [B, D] with 0 and 1
        # select top K highest probabilities, use one hot representation
        preds_oh = select_topk(preds, self.top_k)
        # target_oh [B, D + 1] with 0 and 1
        target_oh = torch.zeros(
            (preds_oh.shape[0], preds_oh.shape[1] + 1),
            device=target.device,
            dtype=torch.int32,
        )
        target = target.long()
        # for undefined targets (-1) use a fake value `num_classes`
        target[target == -1] = self.num_classes
        # fill targets, use one hot representation
        target_oh.scatter_(1, target, 1)
        # target_oh [B, D] (remove the fake target at index `num_classes`)
        target_oh = target_oh[:, :-1]
        # tp [B] with 0 and 1
        tp = (preds_oh * target_oh == 1).sum(dim=1)
        # at least one match between prediction and target
        tp.clip_(max=1)
        # ignore instances where no targets are defined
        mask = target_oh.sum(dim=1) > 0
        tp = tp[mask]
        self.tp.append(tp)  # type: ignore

    def compute(self) -> Tensor:
        """Computes the final metric value.

        Returns:
            Tensor: The computed metric value.
        """
        tp = dim_zero_cat(self.tp)  # type: ignore
        return tp.float().mean()



# MIT License

# Copyright (c) 2022 Intelligent Systems Lab Org

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# File author: Shariq Farooq Bhat

import torch
import torch.nn as nn


@torch.jit.script
def exp_attractor(dx: torch.Tensor, alpha: float = 300, gamma: int = 2) -> torch.Tensor:
    """Calculates the exponential attractor.

    The formula used is:
    dc = exp(-alpha * |dx|^gamma) * dx,
    where dx = a - c, a = attractor point, c = bin center, dc = shift in bin center.

    Args:
        dx (torch.Tensor): The difference tensor dx = Ai - Cj,
                           where Ai is the attractor point and Cj is the bin center.
        alpha (float, optional): Proportional attractor strength.
                                 Determines the absolute strength. Lower alpha = greater attraction.
                                 Defaults to 300.
        gamma (int, optional): Exponential attractor strength.
                               Determines the "region of influence" and indirectly the number of bin centers affected.
                               Lower gamma = farther reach. Defaults to 2.

    Returns:
        torch.Tensor: Delta shifts - dc; New bin centers = Old bin centers + dc.
    """
    return torch.exp(-alpha * (torch.abs(dx) ** gamma)) * (dx)


@torch.jit.script
def inv_attractor(dx: torch.Tensor, alpha: float = 300, gamma: int = 2) -> torch.Tensor:
    """Calculates the inverse attractor.

    The formula used is:
    dc = dx / (1 + alpha * dx^gamma),
    where dx = a - c, a = attractor point, c = bin center, dc = shift in bin center.
    This is the default one according to the accompanying paper.

    Args:
        dx (torch.Tensor): The difference tensor dx = Ai - Cj,
                           where Ai is the attractor point and Cj is the bin center.
        alpha (float, optional): Proportional attractor strength.
                                 Determines the absolute strength. Lower alpha = greater attraction.
                                 Defaults to 300.
        gamma (int, optional): Exponential attractor strength.
                               Determines the "region of influence" and indirectly the number of bin centers affected.
                               Lower gamma = farther reach. Defaults to 2.

    Returns:
        torch.Tensor: Delta shifts - dc; New bin centers = Old bin centers + dc.
    """
    return dx.div(1 + alpha * dx.pow(gamma))


class AttractorLayer(nn.Module):
    """Attractor layer for bin centers.

    This layer computes the new bin centers based on attractor points and previous bin centers.
    The bin centers are bounded on the interval (min_depth, max_depth).
    """

    def __init__(
        self,
        in_features: int,
        n_bins: int,
        n_attractors: int = 16,
        mlp_dim: int = 128,
        min_depth: float = 1e-3,
        max_depth: float = 10,
        alpha: float = 300,
        gamma: int = 2,
        kind: str = "sum",
        attractor_type: str = "exp",
        memory_efficient: bool = False,
    ):
        """Initializes the AttractorLayer.

        Args:
            in_features (int): Number of input features.
            n_bins (int): Number of bins.
            n_attractors (int, optional): Number of attractors. Defaults to 16.
            mlp_dim (int, optional): Dimension of the MLP. Defaults to 128.
            min_depth (float, optional): Minimum depth value. Defaults to 1e-3.
            max_depth (float, optional): Maximum depth value. Defaults to 10.
            alpha (float, optional): Proportional attractor strength. Defaults to 300.
            gamma (int, optional): Exponential attractor strength. Defaults to 2.
            kind (str, optional): Type of operation to perform on delta_c. Defaults to "sum".
            attractor_type (str, optional): Type of attractor to use. Defaults to "exp".
            memory_efficient (bool, optional): Whether to use memory-efficient computation. Defaults to False.
        """
        super().__init__()

        self.n_attractors = n_attractors
        self.n_bins = n_bins
        self.min_depth = min_depth
        self.max_depth = max_depth
        self.alpha = alpha
        self.gamma = gamma
        self.kind = kind
        self.attractor_type = attractor_type
        self.memory_efficient = memory_efficient

        self._net = nn.Sequential(
            nn.Conv2d(in_features, mlp_dim, 1, 1, 0),
            nn.ReLU(inplace=True),
            nn.Conv2d(mlp_dim, n_attractors * 2, 1, 1, 0),  # x2 for linear norm
            nn.ReLU(inplace=True),
        )

    def forward(
        self,
        x: torch.Tensor,
        b_prev: torch.Tensor,
        prev_b_embedding: torch.Tensor = None,
        interpolate: bool = True,
        is_for_query: bool = False,
    ) -> tuple:
        """Forward pass for the AttractorLayer.

        Args:
            x (torch.Tensor): Feature block; shape - (n, c, h, w).
            b_prev (torch.Tensor): Previous bin centers normed; shape - (n, prev_nbins, h, w).
            prev_b_embedding (torch.Tensor, optional): Previous bin embedding. Defaults to None.
            interpolate (bool, optional): Whether to interpolate previous embeddings. Defaults to True.
            is_for_query (bool, optional): Flag indicating if the layer is for query. Defaults to False.

        Returns:
            tuple: New bin centers normed and scaled; shape - (n, nbins, h, w).
        """
        if prev_b_embedding is not None:
            if interpolate:
                prev_b_embedding = nn.functional.interpolate(
                    prev_b_embedding, x.shape[-2:], mode="bilinear", align_corners=True
                )
            x = x + prev_b_embedding

        A = self._net(x)
        eps = 1e-3
        A = A + eps
        n, c, h, w = A.shape
        A = A.view(n, self.n_attractors, 2, h, w)
        A_normed = A / A.sum(dim=2, keepdim=True)  # n, a, 2, h, w
        A_normed = A[:, :, 0, ...]  # n, na, h, w

        b_prev = nn.functional.interpolate(b_prev, (h, w), mode="bilinear", align_corners=True)
        b_centers = b_prev

        if self.attractor_type == "exp":
            dist = exp_attractor
        else:
            dist = inv_attractor

        if not self.memory_efficient:
            func = {"mean": torch.mean, "sum": torch.sum}[self.kind]
            # .shape N, nbins, h, w
            delta_c = func(dist(A_normed.unsqueeze(2) - b_centers.unsqueeze(1)), dim=1)
        else:
            delta_c = torch.zeros_like(b_centers, device=b_centers.device)
            for i in range(self.n_attractors):
                # .shape N, nbins, h, w
                delta_c += dist(A_normed[:, i, ...].unsqueeze(1) - b_centers)

            if self.kind == "mean":
                delta_c = delta_c / self.n_attractors

        b_new_centers = b_centers + delta_c
        B_centers = (self.max_depth - self.min_depth) * b_new_centers + self.min_depth
        B_centers, _ = torch.sort(B_centers, dim=1)
        B_centers = torch.clip(B_centers, self.min_depth, self.max_depth)
        return b_new_centers, B_centers


class AttractorLayerUnnormed(nn.Module):
    """Attractor layer for unbounded bin centers.

    This layer computes the new bin centers based on attractor points and previous bin centers.
    The bin centers are unbounded.
    """

    def __init__(
        self,
        in_features: int,
        n_bins: int,
        n_attractors: int = 16,
        mlp_dim: int = 128,
        min_depth: float = 1e-3,
        max_depth: float = 10,
        alpha: float = 300,
        gamma: int = 2,
        kind: str = "sum",
        attractor_type: str = "exp",
        memory_efficient: bool = False,
    ):
        """Initializes the AttractorLayerUnnormed.

        Args:
            in_features (int): Number of input features.
            n_bins (int): Number of bins.
            n_attractors (int, optional): Number of attractors. Defaults to 16.
            mlp_dim (int, optional): Dimension of the MLP. Defaults to 128.
            min_depth (float, optional): Minimum depth value. Defaults to 1e-3.
            max_depth (float, optional): Maximum depth value. Defaults to 10.
            alpha (float, optional): Proportional attractor strength. Defaults to 300.
            gamma (int, optional): Exponential attractor strength. Defaults to 2.
            kind (str, optional): Type of operation to perform on delta_c. Defaults to "sum".
            attractor_type (str, optional): Type of attractor to use. Defaults to "exp".
            memory_efficient (bool, optional): Whether to use memory-efficient computation. Defaults to False.
        """
        super().__init__()

        self.n_attractors = n_attractors
        self.n_bins = n_bins
        self.min_depth = min_depth
        self.max_depth = max_depth
        self.alpha = alpha
        self.gamma = gamma
        self.kind = kind
        self.attractor_type = attractor_type
        self.memory_efficient = memory_efficient

        self._net = nn.Sequential(
            nn.Conv2d(in_features, mlp_dim, 1, 1, 0),
            nn.ReLU(inplace=True),
            nn.Conv2d(mlp_dim, n_attractors, 1, 1, 0),
            nn.Softplus(),
        )

    def forward(
        self,
        x: torch.Tensor,
        b_prev: torch.Tensor,
        prev_b_embedding: torch.Tensor = None,
        interpolate: bool = True,
        is_for_query: bool = False,
    ) -> tuple:
        """Forward pass for the AttractorLayerUnnormed.

        Args:
            x (torch.Tensor): Feature block; shape - (n, c, h, w).
            b_prev (torch.Tensor): Previous bin centers normed; shape - (n, prev_nbins, h, w).
            prev_b_embedding (torch.Tensor, optional): Previous bin embedding. Defaults to None.
            interpolate (bool, optional): Whether to interpolate previous embeddings. Defaults to True.
            is_for_query (bool, optional): Flag indicating if the layer is for query. Defaults to False.

        Returns:
            tuple: New bin centers unbounded; shape - (n, nbins, h, w). Two outputs just to keep the API consistent with the normed version.
        """
        if prev_b_embedding is not None:
            if interpolate:
                prev_b_embedding = nn.functional.interpolate(
                    prev_b_embedding, x.shape[-2:], mode="bilinear", align_corners=True
                )
            x = x + prev_b_embedding

        A = self._net(x)
        n, c, h, w = A.shape

        b_prev = nn.functional.interpolate(b_prev, (h, w), mode="bilinear", align_corners=True)
        b_centers = b_prev

        if self.attractor_type == "exp":
            dist = exp_attractor
        else:
            dist = inv_attractor

        if not self.memory_efficient:
            func = {"mean": torch.mean, "sum": torch.sum}[self.kind]
            # .shape N, nbins, h, w
            delta_c = func(dist(A.unsqueeze(2) - b_centers.unsqueeze(1)), dim=1)
        else:
            delta_c = torch.zeros_like(b_centers, device=b_centers.device)
            for i in range(self.n_attractors):
                delta_c += dist(A[:, i, ...].unsqueeze(1) - b_centers)  # .shape N, nbins, h, w

            if self.kind == "mean":
                delta_c = delta_c / self.n_attractors

        b_new_centers = b_centers + delta_c
        B_centers = b_new_centers

        return b_new_centers, B_centers



from .ams import AMSDownloader
from .gs import GSDownloader
from .gsv import GSVDownloader
from .kv import KVDownloader
from .mly import MLYDownloader

__all__ = ["AMSDownloader", "GSDownloader", "GSVDownloader", "KVDownloader", "MLYDownloader"]



# Copyright (c) Meta Platforms, Inc. and affiliates.
# All rights reserved.
#
# This source code is licensed under the license found in the
# LICENSE file in the root directory of this source tree.

from .adapters import DatasetWithEnumeratedTargets
from .augmentations import DataAugmentationDINO
from .collate import collate_data_and_cast
from .loaders import SamplerType, make_data_loader, make_dataset
from .masking import MaskingGenerator

__all__ = [
    "DatasetWithEnumeratedTargets",
    "DataAugmentationDINO",
    "collate_data_and_cast",
    "make_data_loader",
    "make_dataset",
    "SamplerType",
    "MaskingGenerator",
]



def infer_type(x):
    """Infer the type of a given input.

    This function attempts to convert a string input into an integer or a float.
    If the input is not a string or cannot be converted, it returns the input as is.

    Args:
        x: The input value to infer the type from. Can be a string, int, float, or any other type.

    Returns:
        The input value converted to int, float, or the original value if conversion is not possible.
    """
    if not isinstance(x, str):
        return x

    try:
        x = int(x)
        return x
    except ValueError:
        pass

    try:
        x = float(x)
        return x
    except ValueError:
        pass

    return x


def parse_unknown(unknown_args):
    """Parse unknown arguments into a dictionary.

    This function takes a list of unknown arguments in the form of strings,
    splits them into key-value pairs, and returns a dictionary with the keys
    stripped of leading dashes and values converted to their inferred types.

    Args:
        unknown_args: A list of strings representing unknown arguments,
                      where each argument can be in the form 'key=value' or just 'key'.

    Returns:
        A dictionary mapping keys to their inferred values.
    """
    clean = []
    for a in unknown_args:
        if "=" in a:
            k, v = a.split("=")
            clean.extend([k, v])
        else:
            clean.append(a)

    keys = clean[::2]
    values = clean[1::2]
    return {k.replace("--", ""): infer_type(v) for k, v in zip(keys, values)}
